/******************************************************************************\
 *                        ____            _ _ 
 *                       |  _ \ ___ _ __ | (_) ___ __ _
 *                       | |_) / _ \ '_ \| | |/ __/ _` |
 *                       |  _ <  __/ |_) | | | (_| (_| |
 *                       |_| \_\___| .__/|_|_|\___\__,_|
 *                                 |_|
 ******************************************************************************
 *                                 
 *  Copyright (c) 2005, Herbert Nachtnebel
 *
 *    based on code from http://www.purists.org [1]
 *
 *  Copyright (c) 2001, Georg Kraml
 *
 *  [1] http://purists.org or http://www.kraml.at (last accessed 2003)
 *      Currently no longer available online and no replacement can be found!
 *      
 ******************************************************************************
 *      
 *  Title:    Replica - File Replication and Backup Utility
 *
 *  Version:  0.1
 *
 *  Date:     01-Nov-2005
 *
 *  Author:   Herbert Nachtnebel (Berri)
 *  
 ******************************************************************************
 *  
 *  Change log:
 *  
 *     Date     Rev  Ver                    Description
 *  ----------- --- ----- ---------------------------------------------------
 *  01-Nov-2005 Ber  0.1  Initial implementation.
 *  
 ******************************************************************************
 *  
 *  Description:
 *
 *    Position Independent AVL Tree for Sorting the Directory Cache
 *    -------------------------------------------------------------
 *
 *    Here, a special AVL Tree for sorting the directory cache is implemented.
 *    All "pointers" in this tree are not realised as real pointers, rather
 *    they are built as offsets from there respective position.  Following
 *    such a "pointer" is realised by adding its value to the address of it.
 *
 *    History
 *    -------
 *
 *    The original code from Georg Kraml is no longer available online, but
 *    it's track can be found in several projects around the globe. Some
 *    examples:
 *    
 *      http://www.pvfs.org/pvfs2/
 *      http://micky.ibh.de/~beck/stuff/nfa/nfapeer.c
 *      http://www.potaroo.net/tools/prefixes/scripts/libpfx.c
 *      
 *    There were no licensing informations given originally, but he explained
 *    the avl algorithm in deep detail at his website and therefore, I think,
 *    it is in the public domain.  It would be nice, if he could approve that. 
 *    
 *    There are basically five changes to the avl tree implementation from
 *    Georg which made the avl tree implementation usable in this environment:
 *
 *    1.) Recursions were eleminated.
 *
 *    2.) The skew of a tree node is not stored in the node itself, but is
 *        held together with the pointer.  Hence, all pointers in the tree
 *        are tagged using the lowest two bits of the pointers.
 *
 *    3.) The remove operator from the tree is optimized.  When deleting a
 *        node, the node which is the immediate predecessor - or the successor,
 *        depending how the subtree is oriented - of the deleted node replaces
 *        it.
 *
 *    4.) As an additional goodie, the tree can be made threaded.  Using this,
 *        finding the successor or predecessor of any node in the tree becomes
 *        stateless.  To turn it on, define the use_avl_threads macro as one.
 *
 *    5.) A rather heavy rewrite changed the representation of the avl node
 *        pointers.  Now, the left and right tree pointers are replaced with
 *        integers which store the offset to the pointed-to node structure in
 *        respect to the current position of the pointer.  Using this trick,
 *        writing out a tree to a file doesn't require any changes to any node
 *        pointers and hence can essentially be made O(1).  Of course, the
 *        write is still linearly dependent on the size of the dataset, but
 *        preparing the write is of constant time.
 *
 *    Probably Georg will no longer recognize his code and for shure will
 *    shake his head if he is ever going to see what I have done to it!
 *    But nevertheless, I am a little bit proud of it, and it is thoroughly
 *    tested on both 32-, and 64-bit machines.  It really should work as is.
 *    If not, _please_ report it.
 *
 *    Implementation Tricks
 *    ---------------------
 *
 *    This implementation requires, that the administered structures added
 *    to a tree contain objects of the type struct avl_node where the tree
 *    pointers are held.  It is not required that these avl_node structures
 *    are placed at specific offsets inside the structure, nor that they
 *    have to be unique.  Using multiple nodes, one is therefore able to bind
 *    a structure to multiple trees.  Consequently, the low level avl_insert
 *    and avl_remove function does not take pointers to the data structures
 *    which are added to the tree, but pointers to the specific avl_node
 *    structure used for searching inside the tree.  Hence, to regain access
 *    to the data structure inside the comparator functions, one has to
 *    recalculate the pointer to the structure from the given avl_node member
 *    pointers by substracting there offset within the structure from there
 *    actual position.  Have a look at the rebase macro and the use of it for
 *    a deeper understanding of the consequencies of this.
 *
 *    Why have I chosen this implementation?  Well, it allows one to write
 *    completely generic low level functions.  Only the basic avl algorithm
 *    has to be implemented by them.  All adjustments to the specific data
 *    layout can then be realized in some simple shell functions around this
 *    low level implementation.  Look at the shell functions at the end of
 *    this file:  everything is thoroughly seperated and layered on top of
 *    each other!
 *    
\******************************************************************************/
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <assert.h>

#include "compat.h"
#include "cache.h"


/******************************************************************************/
/*                                Definitions                                 */
/******************************************************************************/
/*
 *  Even if possible, don't use this with about 2^25 entries for a single tree!
 */
#define AVL_DEPTH 25

/*
 *  We make use of the threads!
 */
#define use_avl_threads 1

/*
 *  Shortcuts for offsets in cache_entry structure.
 */
#define OFFL offsetof(struct avl_node, left)
#define OFFR offsetof(struct avl_node, right)

/*
 *  Given a pointer to a member of a structure, recalculate the pointer
 *  to the structure itself.
 */
#define rebase(type,member,ptr) (type*)(((char*)ptr)-offsetof(type,member))


/******************************************************************************/
/*                                   Types                                    */
/******************************************************************************/
/*
 *  Which of a given node's subtrees is higher?
 *
 *  Note!  These values are added to the tree pointers to tag them.  The
 *         value 0 is reserved for the tree bottom (or threads), but may
 *         not be added to the enumeration since we would get compiler
 *         warnings regarding unhandled enumeration cases in switch
 *         statements.
 */
enum avl_skew
{
    LEFT  = 1,
    NONE  = 2,
    RIGHT = 3
};
#define THREAD 0

/*
 *  Definition of our comparator function.
 */
typedef int (*avl_comparator)(const struct avl_node*, const struct avl_node*);

/*
 *  Definition of our tree walking function.
 */
typedef int (*avl_walker)(struct avl_node *node, void *param, 
                          unsigned int index, unsigned int depth);


/******************************************************************************/
/*                              Inline Functions                              */
/******************************************************************************/
/*
 *  skew
 *  ----
 *
 *    Get the skew of the tagged "pointer"/offset.
 */
inline static enum avl_skew  __attribute__((pure))
skew (avl_root n)
{
    return n & (LEFT | NONE | RIGHT);
}

/*
 *  mask
 *  ----
 *
 *    Get a clean pointer to an avl_node structure.
 */
inline static struct avl_node *  __attribute__((pure))
mask (avl_root *p)
{
    return (struct avl_node *)((char *)p + *p  - skew(*p));
}

/*
 *  none
 *  ----
 *
 *    Convert a tagged offset which is known to be in NONE state into
 *    an avl_node pointer.
 */
inline static struct avl_node *  __attribute__((pure))
none (avl_root *p)
{
    return (struct avl_node *)((char *)p + *p - NONE);
}

/*
 *  assign_tagged
 *  -------------
 *
 *    Concurrently assign and tag a pointer.
 */
inline static void
assign_tagged (avl_root *p, struct avl_node *v, enum avl_skew s)
{
    *p = s + (char*)v - (char*)p;
}

/*
 *  followed_left
 *  -------------
 *
 *    Check if we followed the left pointer of the node.
 */
inline static int  __attribute__((pure))
followed_left (avl_root *p, avl_root *n)
{
    return (char*)p < (char*)n + *n;
}

/*
 *  follow_thread
 *  -------------
 *
 *    Follow a tagged pointer which is known to be in THREAD state.
 */
inline static void *  __attribute__((pure))
follow_thread (avl_root *p)
{
    return (char *)p + *p;
}


/******************************************************************************/
/*                         Low Level AVL Tree Functions                       */
/******************************************************************************/
/*
 *  avl_left_grown
 *  --------------
 *
 *    Helper function for inserting into AVL tree.
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node.  This node's left subtree
 *                has just grown due to item insertion; its "skew" flag needs
 *                adjustment, and the local tree (the subtree of which this
 *                node is the root node) may have become unbalanced.
 *
 *  Return values:
 *
 *    true        The local tree could be rebalanced or was balanced from the
 *                start.  The parent that called this function may assume the
 *                entire tree is valid.
 *
 *    false       The local tree was balanced, but has grown in height.
 *                Do not assume the entire tree is valid.
 */
inline static int
avl_left_grown (avl_root *n)
{
    avl_root B, C, *p, *pn;

    switch (skew(*n))
    {
        case LEFT:
            *n += NONE - LEFT;
            if (LEFT == skew(none(n)->left))
            {
                none(n)->left += NONE - LEFT;
            }
            else
            {
                none(n)->left += NONE - RIGHT;
                p = &none(n)->left;
                switch (skew(none(p)->right))
                {
                    case LEFT:
                        none(p)->right += NONE - LEFT;

                        pn = &none(p)->right;  B = *pn - NONE + OFFR;
                        *p += B;     C = none(p)->left - NONE + OFFL;
                        none(p)->left = -B + NONE - OFFL;
                        *pn += C;

                        pn = &none(n)->left;   B = *pn - NONE + OFFL;
                        *n += B;    C = none(n)->right - NONE + OFFR;
                        none(n)->right = -B + RIGHT - OFFR;
                        goto second_thread;

                    case RIGHT:
                        none(p)->right += NONE - RIGHT;

                        pn = &none(p)->right;  B = *pn - NONE + OFFR;
                        *p += B;     C = none(p)->left - NONE + OFFL;
                        none(p)->left = -B + LEFT - OFFL;
                        goto first_thread;

                    case NONE:
                        pn = &none(p)->right;  B = *pn - NONE + OFFR;
                        *p += B;     C = none(p)->left - NONE + OFFL;
                        none(p)->left = -B + NONE - OFFL;

                    first_thread:
                        *pn += C;
                        if (use_avl_threads)
                            if (*pn + OFFR == 0)
                                *pn = -C - NONE - OFFR;
                        break;
                }
            }
            pn = &none(n)->left;   B = *pn - NONE + OFFL;
            *n += B;    C = none(n)->right - NONE + OFFR;
            none(n)->right = -B + NONE - OFFR;

        second_thread:
            *pn += C;
            if (use_avl_threads)
                if (*pn + OFFL == 0)
                    *pn = -C - NONE - OFFL;
            break;

        case RIGHT:
            *n += NONE - RIGHT;
            break;

        case NONE:
            *n += LEFT - NONE;
            return 0;
    }
    return 1;
}

/*
 *  avl_right_grown
 *  ---------------
 *
 *    Helper function for inserting into AVL tree.
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node.  This node's right subtree
 *                has just grown due to item insertion; its "skew" flag needs
 *                adjustment, and the local tree (the subtree of which this
 *                node is the root node) may have become unbalanced.
 *
 *  Return values:
 *
 *    true        The local tree could be rebalanced or was balanced from the
 *                start.  The parent that called this function may assume the
 *                entire tree is valid.
 *
 *    false       The local tree was balanced, but has grown in height.
 *                Do not assume the entire tree is valid.
 */
inline static int
avl_right_grown (avl_root *n)
{
    avl_root B, C, *p, *pn;

    switch (skew(*n))
    {
        case RIGHT:
            *n += NONE - RIGHT;
            if (RIGHT == skew(none(n)->right))
            {
                none(n)->right += NONE - RIGHT;
            }
            else
            {
                none(n)->right += NONE-LEFT;
                p = &none(n)->right;
                switch (skew(none(p)->left))
                {
                    case RIGHT:
                        none(p)->left += NONE - RIGHT;

                        pn = &none(p)->left;   B = *pn - NONE + OFFL;
                        *p += B;    C = none(p)->right - NONE + OFFR;
                        none(p)->right = -B + NONE - OFFR;
                        *pn += C;

                        pn = &none(n)->right;  B = *pn - NONE + OFFR;
                        *n += B;     C = none(n)->left - NONE + OFFL;
                        none(n)->left = -B + LEFT - OFFL;
                        goto second_thread;

                    case LEFT:
                        none(p)->left += NONE - LEFT;

                        pn = &none(p)->left;   B = *pn - NONE + OFFL;
                        *p += B;    C = none(p)->right - NONE + OFFR;
                        none(p)->right = -B + RIGHT - OFFR;
                        goto first_thread;

                    case NONE:
                        pn = &none(p)->left;   B = *pn - NONE + OFFL;
                        *p += B;    C = none(p)->right - NONE + OFFR;
                        none(p)->right = -B + NONE - OFFR;

                    first_thread:
                        *pn += C;
                        if (use_avl_threads)
                            if (*pn + OFFL == 0)
                                *pn = -C - NONE - OFFL;
                        break;
                }
            }
            pn = &none(n)->right;  B = *pn - NONE + OFFR;
            *n += B;     C = none(n)->left - NONE + OFFL;
            none(n)->left = -B + NONE - OFFL;

        second_thread:
            *pn += C;
            if (use_avl_threads)
                if (*pn + OFFR == 0)
                    *pn = -C - NONE - OFFR;
            break;

        case LEFT:
            *n += NONE - LEFT;
            break;

        case NONE:
            *n += RIGHT - NONE;
            return 0;
    }
    return 1;
}

/*
 *  avl_insert
 *  ----------
 *
 *    AVL tree insertion operation.  The key is added to the tree if there
 *    is no identical one registered before.  In sharp contrast to other
 *    implementations, the key has to be a prebuilt object here.
 *
 *    Checking whether the key was really added or existed already can be
 *    done by comparing the returned pointer to the given key.  When they
 *    are equal, the key was added.
 *
 *    The members of the tree are sorted in ascending order according to a
 *    comparison function, which is called with two arguments that point
 *    to the struct avl_node members of the objects being compared.
 * 
 *  Parameters:
 *
 *    n           Address of a pointer to a node.  It points to the root
 *                node of the avl tree where the item is inserted.
 *
 *    comp        Comparision function which is used to sort the tree. 
 *                It must return an integer less than, equal to, or greater
 *                than zero if the first argument is considered to be less
 *                than, equal to, or greater than the second, respectively.
 *
 *    key         Pointer to a key which should be bound to the tree.
 *                This pointer may not be NULL!
 *
 *  Return values:
 *
 *    non-null    Pointer to the node structure for the given key.  If it
 *                existed before, the pointer to the existing node structure
 *                is returned.  Else, the pointer to the given key is
 *                returned.
 */
static struct avl_node *
avl_insert (avl_root *n, avl_comparator comp, struct avl_node *key)
{
    unsigned int sp = 0;		// stack pointer
    avl_root *stack[AVL_DEPTH];		// tree stack

    struct avl_node *f = f;
    int c = 0;
    
    // traverse tree down
    while (skew(*n))
    {
        f = mask(n);
        c = comp (key, f);
        
        stack[sp++] = n;
        if (c < 0)
            n = &f->left;
        else if (c > 0)
            n = &f->right;
        else
            return f;
    }
    
    // insert the given new node
    key->left = key->right = 0;
    if (use_avl_threads)
    {
        if (c < 0)
        {
            assign_tagged (&key->right, f, THREAD);
            if (*n)
                assign_tagged (&key->left, mask(n), THREAD);
        }
        else if (c > 0)
        {
            assign_tagged (&key->left, f, THREAD);
            if (*n)
                assign_tagged (&key->right, mask(n), THREAD);
        }
    }
    assign_tagged (n, key, NONE);

    // traverse tree up, balancing it as long as needed
    while (sp)
    {
        c = followed_left (n, stack[--sp]);
        n = stack[sp];
        if (c ? avl_left_grown(n) : avl_right_grown(n))
            break;
    }
    return key;
}


/*
 *  avl_left_shrunk
 *  ---------------
 *
 *    Helper function for deleting from AVL tree.
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node.  The node's left subtree
 *                has just shrunk due to item removal; its "skew" flag needs
 *                adjustment, and the local tree (the subtree of which this
 *                node is the root node) may have become unbalanced.
 *
 *  Return values:
 *
 *    true        The parent that called this function may assume that the
 *                entire tree is valid.
 *
 *    false       Do not assume the entire tree is valid.
 */                
inline static int
avl_left_shrunk (avl_root *n)
{
    avl_root B, C, *p, *pn;

    switch (skew(*n))
    {
        case RIGHT:
            *n += NONE - RIGHT;
            switch (skew(none(n)->right))
            {
                case RIGHT:
                    none(n)->right += NONE - RIGHT;
                    break;

                case LEFT:
                    none(n)->right += NONE - LEFT;
                    p = &none(n)->right;
                    switch (skew(none(p)->left))
                    {
                        case RIGHT:
                            none(p)->left += NONE - RIGHT;

                            pn = &none(p)->left;   B = *pn - NONE + OFFL;
                            *p += B;    C = none(p)->right - NONE + OFFR;
                            none(p)->right = -B + NONE - OFFR;
                            *pn += C;

                            pn = &none(n)->right;  B = *pn - NONE + OFFR;
                            *n += B;     C = none(n)->left - NONE + OFFL;
                            none(n)->left = -B + LEFT - OFFL;
                            goto second_thread;

                        case LEFT:
                            none(p)->left += NONE - LEFT;

                            pn = &none(p)->left;   B = *pn - NONE + OFFL;
                            *p += B;    C = none(p)->right - NONE + OFFR;
                            none(p)->right = -B + RIGHT - OFFR;
                            goto first_thread;

                        case NONE:
                            pn = &none(p)->left;   B = *pn - NONE + OFFL;
                            *p += B;    C = none(p)->right - NONE + OFFR;
                            none(p)->right = -B + NONE - OFFR;

                        first_thread:
                            *pn += C;
                            if (use_avl_threads)
                                if (*pn + OFFL == 0)
                                    *pn = -C - NONE - OFFL;
                            break;
                    }
                    break;

                case NONE:
                    pn = &none(n)->right;  B = *pn - NONE + OFFR;
                    *n += B;     C = none(n)->left - NONE + OFFL;
                    none(n)->left = -B + RIGHT - OFFL;
                    *pn += C;
                    *n += LEFT - NONE;
                    return 1;
            }
            pn = &none(n)->right;   B = *pn - NONE + OFFR;
            *n += B;      C = none(n)->left - NONE + OFFL;
            none(n)->left = -B + NONE - OFFL;

        second_thread:
            *pn += C;
            if (use_avl_threads)
                if (*pn + OFFR == 0)
                    *pn = -C - NONE - OFFR;
            break;

        case LEFT:
            *n += NONE - LEFT;
            break;

        case NONE:
            *n += RIGHT - NONE;
            return 1;
    }
    return 0;
}

/*
 *  avl_right_shrunk
 *  ----------------
 *
 *    Helper function for deleting from AVL tree.
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node.  The node's right subtree
 *                has just shrunk due to item removal; its "skew" flag needs
 *                adjustment, and the local tree (the subtree of which this
 *                node is the root node) may have become unbalanced.
 *
 *  Return values:
 *
 *    true        The parent that called this function may assume that the
 *                entire tree is valid.
 *
 *    false       Do not assume the entire tree is valid.
 */                
inline static int
avl_right_shrunk (avl_root *n)
{
    avl_root B, C, *p, *pn;

    switch (skew(*n))
    {
        case LEFT:
            *n += NONE - LEFT;
            switch (skew(none(n)->left))
            {
                case LEFT:
                    none(n)->left += NONE - LEFT;
                    break;

                case RIGHT:
                    none(n)->left += NONE - RIGHT;
                    p = &none(n)->left;
                    switch (skew(none(p)->right))
                    {
                        case LEFT:
                            none(p)->right += NONE-LEFT;

                            pn = &none(p)->right;  B = *pn - NONE + OFFR;
                            *p += B;     C = none(p)->left - NONE + OFFL;
                            none(p)->left = -B + NONE - OFFL;
                            *pn += C;

                            pn = &none(n)->left;   B = *pn - NONE + OFFL;
                            *n += B;    C = none(n)->right - NONE + OFFR;
                            none(n)->right = -B + RIGHT - OFFR;
                            goto second_thread;

                        case RIGHT:
                            none(p)->right += NONE - RIGHT;

                            pn = &none(p)->right;  B = *pn - NONE + OFFR;
                            *p += B;     C = none(p)->left - NONE + OFFL;
                            none(p)->left = -B + LEFT - OFFL;
                            goto first_thread;

                        case NONE:
                            pn = &none(p)->right;  B = *pn - NONE + OFFR;
                            *p += B;     C = none(p)->left - NONE + OFFL;
                            none(p)->left = -B + NONE - OFFL;

                        first_thread:
                            *pn += C;
                            if (use_avl_threads)
                                if (*pn + OFFR == 0)
                                    *pn = -C - NONE - OFFR;
                            break;
                    }
                    break;

                case NONE:
                    pn = &none(n)->left;   B = *pn - NONE + OFFL;
                    *n += B;    C = none(n)->right - NONE + OFFR;
                    none(n)->right = -B + LEFT - OFFR;
                    *pn += C;
                    *n += RIGHT - NONE;
                    return 1;
            }
            pn = &none(n)->left;   B = *pn - NONE + OFFL;
            *n += B;    C = none(n)->right - NONE + OFFR;
            none(n)->right = -B + NONE - OFFR;

        second_thread:
            *pn += C;
            if (use_avl_threads)
                if (*pn + OFFL == 0)
                    *pn = -C - NONE - OFFL;
            break;

        case RIGHT:
            *n += NONE - RIGHT;
            break;

        case NONE:
            *n += LEFT - NONE;
            return 1;
    }
    return 0;
}

/*
 *  avl_remove
 *  ----------
 *
 *    Remove an item from the AVL tree.
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node.  It points to the root node
 *                of the avl tree from which the item is deleted.
 *
 *    comp        Comparision function which is used to sort the tree. 
 *                It must return an integer less than, equal to, or greater
 *                than zero if the first argument is considered to be less
 *                than, equal to, or greater than the second, respectively.
 *
 *    key         Pointer to the key of item which is going to be removed.
 *                Since the tree is traversed to the object which will be
 *                deleted, it is not necessary that the pointer to the real
 *                object has to be given.  Instead, a partially constructed
 *                key element can be given since only the - user supplied -
 *                comparision function uses any fields from this key.
 *
 *  Return values:
 *
 *    non-null    The item has been removed.
 *
 *    null        The tree does not contain an item yielding the key value
 *                provided by the caller.
 */
static struct avl_node *
avl_remove (avl_root *n, avl_comparator comp, struct avl_node *key)
{
    unsigned int sp = 0, relink;	// stack pointer
    avl_root *stack[AVL_DEPTH];		// tree stack

    struct avl_node *v = v;		// victim replacing the deleted node
    struct avl_node *f;			// node to delete
    int c;

    // traverse tree down
    while (skew(*n))
    {
        f = mask(n);
        c = comp (key, f);
        
        stack[sp++] = n;
        if (c < 0)
            n = &f->left;
        else if (c > 0)
            n = &f->right;
        else
            goto delete;
    }
    return NULL;

  delete:
    relink = sp;
    c = (skew(*n) - NONE) | skew(f->left);
    if (c < 0)
    {
        // search for victim (the predecessor of the deleted node)
        n = &f->left;
        do {
            stack[sp++] = n;
            v = mask(n);
            n = &v->right;
        } while (skew(*n));

        // adjust other thread (in successor node, opposite of victim)
        if (use_avl_threads)
        {
            n = &f->right;
            if (skew(*n))
            {
                do n = &mask(n)->left; while (skew(*n));
                assign_tagged (n, v, THREAD);
            }
        }

        // disconnect victim
        n = stack[--sp];
        *n += v->left + OFFL - skew(*n);
                
        // if the victim is a leaf node, the link to it becomes a thread
        if (use_avl_threads)
            if (*n + OFFR == 0)
                assign_tagged (n, v, THREAD);
    }
    else if (c > 0)
    {
        // search for victim (the successor of the deleted node)
        n = &f->right;
        do {
            stack[sp++] = n;
            v = mask(n);
            n = &v->left;
        } while (skew(*n));

        // adjust other thread (in predecessor node, opposite of victim)
        if (use_avl_threads)
        {
            n = &f->left;
            if (skew(*n))
            {
                do n = &mask(n)->right; while (skew(*n));
                assign_tagged (n, v, THREAD);
            }
        }

        // disconnect victim
        n = stack[--sp];
        *n += v->right + OFFR - skew(*n);

        // if the victim is a leaf node, the link to it becomes a thread
        if (use_avl_threads)
            if (*n + OFFL == 0)
                assign_tagged (n, v, THREAD);
    }
    else
    {
        // the deleted item sits at the tree bottom, no threads point to it
        relink = 0;

        // dropping the deleted node, the original pointer to this item gets
        // changed into a thread to the predecessor or successor, respectively
        n = stack[--sp];
        if (n == follow_thread (&f->right))
            *n = use_avl_threads && f->left  ? *n + f->left  + OFFL - NONE : 0;
        else
            *n = use_avl_threads && f->right ? *n + f->right + OFFR - NONE : 0;
    }

    // traverse tree up, balancing it as long as needed
    while (sp)
    {
        c = followed_left (n, stack[--sp]);
        n = stack[sp];
        if (c ? avl_left_shrunk(n) : avl_right_shrunk(n))
            break;
    }

    // the tree is in shape again, now relink the victim to the tree
    if (relink)
    {
        c = (f-v) * sizeof *f;
        n = stack[--relink];
        v->left  = f->left  ? f->left  + c : 0;
        v->right = f->right ? f->right + c : 0;
        *n -= c;
    }

    // finally, signal that it is save to delete the entry
    f->left = f->right = 0;
    return f;
}


/*
 *  avl_find
 *  --------
 *
 *    Retrieve the datum corresponding to a given key.
 *
 *  Parameters:
 *
 *    n           Address of the tree root pointer.                         
 *
 *    comp        Comparision function which is used to find the object.
 *                It must return an integer less than, equal to, or greater
 *                than zero if the first argument is considered to be less
 *                than, equal to, or greater than the second, respectively.
 *
 *    key         Address of key of item to be accessed.
 *
 *  Return values:
 *
 *    non-null    An item yielding the key provided has been found, the return
 *                value points to the node structure attached to it.
 *
 *    null        The item could not be found.
 */
static struct avl_node *
avl_find (avl_root *n, avl_comparator comp, const struct avl_node *key)
{
    while (skew(*n))
    {
        struct avl_node *p = mask(n);
        int c = comp (key, p);

        if (c < 0)
            n = &p->left;
        else if (c > 0)
            n = &p->right;
        else
            return p;
    }
    return NULL;
}

/*
 *  avl_head
 *  --------
 *
 *    Retrieve the datum corresponding to a given key. If the node couldn't
 *    be found return the tree node which is the predecessor of the given key.
 *
 *  Parameters:
 *
 *    n           Address of the tree root pointer.                         
 *
 *    comp        Comparision function which is used to traverse the tree.
 *                It must return an integer less than, equal to, or greater
 *                than zero if the first argument is considered to be less
 *                than, equal to, or greater than the second, respectively.
 *
 *    key         Address of key of item to be accessed.
 *
 *  Return values:
 *
 *    non-null    An item yielding the key provided has been found, the
 *                return value points to the node structure attached to it.
 *
 *    null        The key is less than the key of the first item in tree.
 */
static struct avl_node *
avl_head (avl_root *n, avl_comparator comp, const struct avl_node *key)
{
    struct avl_node *m = NULL;
    
    while (skew(*n))
    {
        struct avl_node *p = mask(n);
        int c = comp (key, p);

        if (c < 0)
            n = &p->left;
        else if (c > 0)
        {
            m = p;
            n = &p->right;
        }
        else
            return p;
    }
    return m;
}

/*
 *  avl_pred
 *  --------
 *
 *    Retrieve the tree item which is the predecessor of the given key.
 *
 *  Parameters:
 *
 *    n           Address of the tree root pointer.                         
 *
 *    comp        Comparision function which is used to traverse the tree.
 *                It must return an integer less than, equal to, or greater
 *                than zero if the first argument is considered to be less
 *                than, equal to, or greater than the second, respectively.
 *
 *    key         Key of item to be accessed.
 *
 *  Return values:
 *
 *    non-null    An item yielding the key provided has been found, the
 *                return value points to the node structure attached to it.
 *
 *    null        There exists no item before the refered key.
 */
static struct avl_node *
avl_pred (avl_root *n, avl_comparator comp, const struct avl_node *key)
{
    struct avl_node *m = NULL;

    while (skew(*n))
    {
        struct avl_node *p = mask(n);
        int c = comp (key, p);

        if (c <= 0)
            n = &p->left;
        else
        {
            m = p;
            n = &p->right;
        }
    }
    return m;
}

/*
 *  avl_succ
 *  --------
 *
 *    Retrieve the tree item which is the successor of the given key.
 *
 *  Parameters:
 *
 *    n           Address of the tree root pointer.                         
 *
 *    comp        Comparision function which is used to traverse the tree.
 *                It must return an integer less than, equal to, or greater
 *                than zero if the first argument is considered to be less
 *                than, equal to, or greater than the second, respectively.
 *
 *    key         Address of key of item to be accessed.
 *
 *  Return values:
 *
 *    non-null    An item yielding the key provided has been found, the
 *                return value points to the node structure attached to it.
 *
 *    null        There exists no item after the refered key.
 */
static struct avl_node *
avl_succ (avl_root *n, avl_comparator comp, const struct avl_node *key)
{
    struct avl_node *m = NULL;

    while (skew(*n))
    {
        struct avl_node *p = mask(n);
        int c = comp (key, p);
        
        if (c >= 0)
            n = &p->right;
        else
        {
            m = p;
            n = &p->left;
        }
    }
    return m;
}

/*
 *  avl_first
 *  ---------
 *
 *    Retrieve the first tree item in AVL tree.
 *
 *  Parameters:
 *
 *    n           Address of the tree root pointer.                         
 *
 *  Return values:
 *
 *    non-null    The pointer to the first node structure in tree.
 *
 *    null        The tree contains no items.
 */
static struct avl_node *
avl_first (avl_root *n)
{
    struct avl_node *p = NULL;
    
    while (skew(*n))
    {
        p = mask(n);
        n = &p->left;
    }
    return p;
}

/*
 *  avl_last
 *  --------
 *
 *    Retrieve the last tree item in AVL tree.
 *
 *  Parameters:
 *
 *    n           Address of the tree root pointer.                         
 *
 *  Return values:
 *
 *    non-null    The pointer to the last node structure in tree.
 *
 *    null        The tree contains no items.
 */
static struct avl_node *
avl_last (avl_root *n)
{
    struct avl_node *p = NULL;
    
    while (skew(*n))
    {
        p = mask(n);
        n = &p->right;
    }
    return p;
}

/*
 *  avl_prev
 *  --------
 *
 *    Retrieve the tree item which is the predecessor of the given tree node.
 *    This variant makes use of the threads in the tree to find the previous
 *    node and therefore doesn't need the comparator function.
 *
 *  Parameters:
 *
 *    n           Address of the tree root pointer.                         
 *
 *  Return values:
 *
 *    non-null    Pointer to previous tree node.
 *
 *    null        The tree node is the first one in tree.
 */
static struct avl_node *
avl_prev (struct avl_node *node)
{
    assert(use_avl_threads);
    
    if (node)
    {
        avl_root *n = &node->left;
        struct avl_node *p = *n ? mask(n) : 0;

        if (skew(*n))
        {
            while (skew(p->right))
            {
                n = &p->right;
                p = mask(n);
            }
        }
        return p;
    }
    return NULL;
}

/*
 *  avl_next
 *  --------
 *
 *    Retrieve the tree item which is the successor of the given tree node.
 *    This variant makes use of the threads in the tree to find the next
 *    node and therefore doesn't need the comparator function.
 *
 *  Parameters:
 *
 *    n           Address of the tree root pointer.                         
 *
 *  Return values:
 *
 *    non-null    Pointer to the next tree node.
 *
 *    null        The tree node is the first one in tree.
 */
static struct avl_node *
avl_next (struct avl_node *node)
{
    assert(use_avl_threads);

    if (node)
    {
        avl_root *n = &node->right;
        struct avl_node *p = *n ? mask(n) : 0;

        if (skew(*n))
        {
            while (skew(p->left))
            {
                n = &p->left;
                p = mask(n);
            }
        }
        return p;
    }
    return NULL;
}

/*
 *  avl_walk
 *  --------
 *
 *    Tree traversal in ascending order.
 *
 *  Parameters:
 *
 *    n           Address of the tree root pointer.                         
 *
 *    function    Worker function to be called for every node.
 *
 *    param       Arbitrary parameter given to worker function.
 *
 *  Return values:
 *
 *    zero or     The whole tree was traversed.  The return value is the
 *    positive    sum of all values returned by the worker functions.
 *
 *    negative    The tree traversal was aborted due to a worker function
 *                returning a value less than zero.
 */
static int
avl_walk (avl_root *n, avl_walker function, void *param)
{
    int rv, sum = 0;
    unsigned int cnt = 0;
    unsigned int sp  = 0;
    avl_root *stack[AVL_DEPTH];
    struct avl_node *p;
    
    while (1)
    {
        while (skew(*n))
        {
            stack[sp++] = n;
            n = &mask(n)->left;
        }

        if (!sp)
            return sum;
        
        p = mask(stack[--sp]);
        n = &p->right;
        
        if ((rv = function(p, param, cnt++, sp)) < 0)
            return rv;
        sum += rv;
    }
}

/*
 *  avl_reverse_walk
 *  ----------------
 *
 *    Tree traversal in descending order.
 *
 *  Parameters:
 *
 *    n           Address of the tree root pointer.                         
 *
 *    function    Worker function to be called for every node.
 *
 *    param       Arbitrary parameter given to worker function.
 *
 *  Return values:
 *
 *    zero or     The whole tree was traversed.  The return value is the
 *    positive    sum of all values returned by the worker functions.
 *
 *    negative    The tree traversal was aborted due to a worker function
 *                returning a value less than zero.
 */
static int
avl_reverse_walk (avl_root *n, avl_walker function, void *param)
{
    int rv, sum = 0;
    unsigned int cnt = 0;
    unsigned int sp  = 0;
    avl_root *stack[AVL_DEPTH];
    struct avl_node *p;
    
    while (1)
    {
        while (skew(*n))
        {
            stack[sp++] = n;
            n = &mask(n)->right;
        }

        if (!sp)
            return sum;
        
        p = mask(stack[--sp]);
        n = &p->left;
        
        if ((rv = function(p, param, cnt++, sp)) < 0)
            return rv;
        sum += rv;
    }
}

/******************************************************************************/
/*             Shell Functions Around the AVL Tree Implementation             */
/******************************************************************************/

/*
 *  ce_sort_name
 *  ------------
 *
 *    Cache entries are sorted by name.
 */
static int
ce_sort_name (const struct avl_node *a, const struct avl_node *b)
{
    const struct cache_entry *A = rebase(struct cache_entry, ce_neighbors, a);
    const struct cache_entry *B = rebase(struct cache_entry, ce_neighbors, b);
    
    return strcmp (A->ce_name, B->ce_name);
}

/*
 *  find_first_entry
 *  ----------------
 *
 *    Get the first cache entry of the given tree.
 */
struct cache_entry *
find_first_entry (avl_root *root)
{
    struct cache_entry *rv = NULL;

    if (root)
        rv = rebase( struct cache_entry, ce_neighbors,
                     avl_first(root) );
    return rv;
}

/*
 *  find_next_entry
 *  ---------------
 *
 *    Given a cache entry in a tree, get the following cache entry.
 */
struct cache_entry *
find_next_entry (avl_root *root, struct cache_entry *entry)
{
    struct cache_entry *rv = NULL;

    if (root && entry)
    {
        if (use_avl_threads)
            rv = rebase( struct cache_entry, ce_neighbors,
                         avl_next(&entry->ce_neighbors) );
        else
            rv = rebase( struct cache_entry, ce_neighbors,
                         avl_succ(root, ce_sort_name, &entry->ce_neighbors) );

        /*
         *  A very simple check to find implementation errors: check that
         *  both implementations of the element successor operation match.
         */
        if (use_avl_threads)
            assert( rebase( struct cache_entry, ce_neighbors,
                            avl_succ(root, ce_sort_name, &entry->ce_neighbors) ) ==
                    rv );
    }
    return rv;
}

/*
 *  find_entry
 *  ----------
 *
 *    Search in a tree for a cache entry given the name of it.
 */
struct cache_entry *
find_entry (avl_root *root, const char *name)
{
    /*
     *  Note!  For searching we need the temporary cache entry structure "key".
     *         But since the algorithm doesn't look at other entries beside
     *         the sorting key, we just undo the pointer arithmetic which will
     *         be done in the ce_sort_name function above to find the name.
     */
    struct cache_entry *rv = NULL;
    struct cache_entry *key = rebase (struct cache_entry, ce_name, name);

    if (root)
        rv = rebase( struct cache_entry, ce_neighbors,
                     avl_find(root, ce_sort_name, &key->ce_neighbors) );
    return rv;
}

/*
 *  link_entry
 *  ----------
 *
 *    Add a cache entry to a tree.
 */
struct cache_entry *
link_entry (avl_root *root, struct cache_entry *key)
{
    struct cache_entry *rv = NULL;

    if (root && key)
        rv = rebase( struct cache_entry, ce_neighbors,
                     avl_insert(root, ce_sort_name, &key->ce_neighbors) );
    return rv;
}

/*
 *  unlink_entry
 *  ------------
 *
 *    Remove a cache entry from a tree.
 */
int
unlink_entry (avl_root *root, struct cache_entry *key)
{
    return avl_remove(root, ce_sort_name, &key->ce_neighbors) != NULL;
}

/******************************************************************************/

struct walk_context
{
    int (*walker)(struct cache_entry *, void *);
    void *data;
};  

static int
all_entries (struct avl_node *node, void *param,
             unsigned int index, unsigned int depth)
{
    struct walk_context *context = param;
    struct cache_entry *ce = rebase (struct cache_entry, ce_neighbors, node);

    return (*context->walker)(ce, context->data);
}

int
walk_cache_entries (avl_root *root,
                    int (*walker)(struct cache_entry *, void *),
                    void *data)
{
    struct walk_context context = { walker, data };
    
    return root ? avl_walk (root, all_entries, &context) : 0;
}

/******************************************************************************/

/*
 *  A freed cache entry is linked to the cache_header as this element.
 */
struct cache_free
{
    avl_node _i_am_forbidden_;          /* overlayed over ce_neighbors */
    avl_node cf_addr, cf_size;          /* sorted neighbor entries */
    size_t length;                      /* length of free block */
};

/*
 *  cf_sort_addr
 *  ------------
 *
 *    Sort freed cache entries by it's address.
 */
static int
cf_sort_addr (const struct avl_node *a, const struct avl_node *b)
{
    if (a < b)
        return -1;
    if (a > b)
        return 1;
    return 0;
}

/*
 *  cf_sort_size
 *  ------------
 *
 *    Sort freed cache entries by it's size and it's address.
 */
static int
cf_sort_size (const struct avl_node *a, const struct avl_node *b)
{
    const struct cache_free *A = rebase (struct cache_free, cf_size, a);
    const struct cache_free *B = rebase (struct cache_free, cf_size, b);

    if (A->length > B->length)
        return -1;
    if (A->length < B->length)
        return 1;
    if (a < b)
        return -1;
    if (a > b)
        return 1;
    return 0;
}

/*
 *  pull_free_entry
 *  ---------------
 *
 *    Take a prevously freed cache entry from the free "list".
 */
void
pull_free_entry (struct cache_header *hdr, struct cache_entry *new, size_t size)
{
    struct cache_free *cf = (struct cache_free*) new;
    size_t length = cf->length;

    /*
     *  Remove object from both free trees.
     */
    avl_remove (&hdr->hdr_free_addr, cf_sort_addr, &cf->cf_addr);
    avl_remove (&hdr->hdr_free_size, cf_sort_size, &cf->cf_size);
    cf->length = 0;

    /*
     *  If there is some space left, re-add it.
     */
    if (length > size)
    {
        cf = (struct cache_free*)((char*)new + size);
        cf->length = length - size;
        avl_insert (&hdr->hdr_free_addr, cf_sort_addr, &cf->cf_addr);
        avl_insert (&hdr->hdr_free_size, cf_sort_size, &cf->cf_size);
    }
    else
    {
        assert (length == size);
    }
}

/*
 *  push_free_entry
 *  ---------------
 *
 *    Free a cache entry by pushing it to the free "list".
 */
void
push_free_entry (struct cache_header *hdr, struct cache_entry *old, size_t size)
{
    struct cache_free *cf = (struct cache_free *) old;
    struct avl_node *prev, *next;

    /*
     *  Clear the old contents.
     */
    memset (old, 0, size);

    /*
     *  Get the neighboring objects, if any.
     */
    prev = avl_head (&hdr->hdr_free_addr, cf_sort_addr, &cf->cf_addr);
    next = prev ? avl_next (prev) : avl_first (&hdr->hdr_free_addr);

    /*
     *  Can we join this freed object with the following one?
     */
    if (next)
    {
        struct cache_free *succ = rebase (struct cache_free, cf_addr, next);

        if ((char*)cf + size == (char*)succ)
        {
            /*
             *  We generate a new freed object, hence drop the old one
             */
            size += succ->length;
            avl_remove (&hdr->hdr_free_addr, cf_sort_addr, &succ->cf_addr);
            avl_remove (&hdr->hdr_free_size, cf_sort_size, &succ->cf_size);
            succ->length = 0;
        }
    }

    /*
     *  And what about the previous object?
     */
    if (prev)
    {
        struct cache_free *pred = rebase (struct cache_free, cf_addr, prev);

        if ((char*)pred + pred->length == (char*)cf)
        {
            /*
             *  We are going to change the size, therefore temporarily remove
             *  it from the sized tree and re-add it again with the new size.
             */
            avl_remove (&hdr->hdr_free_size, cf_sort_size, &pred->cf_size);
            pred->length += size;
            avl_insert (&hdr->hdr_free_size, cf_sort_size, &pred->cf_size);
            return;
        }
    }    

    /*
     *  Well, that's it, add this new freed object to the free trees.
     */
    cf->length = size;
    avl_insert (&hdr->hdr_free_addr, cf_sort_addr, &cf->cf_addr);
    avl_insert (&hdr->hdr_free_size, cf_sort_size, &cf->cf_size);
}

/*
 *  locate_free_entry
 *  -----------------
 *
 *    Try to reuse a prevously freed cache entry.
 */
struct cache_entry *
locate_free_entry (struct cache_header *hdr, avl_root *hint, size_t size)
{
    struct cache_entry *ce = NULL;
    avl_root *n = &hdr->hdr_free_size;
    
    while (skew(*n))
    {
        struct avl_node *p = mask(n);
        struct cache_free *f = rebase (struct cache_free, cf_size, p);
        int c = f->length - size;

        if (!c)
            c = (char*)hint - (char*)f;
        if (c < 0)
            n = &p->left;
        else if (c > 0)
        {
            ce = (struct cache_entry *) f;
            n = &p->right;
        }
    }

    /*
     *  Note!  We also should check if larger objects are better positioned
     *         relative to hint.
     */
    return ce;
}

/******************************************************************************/

#include <stdio.h>
#include "index.h"

static int dump_addr_entry();
static int dump_size_entry();

int
dump_freed (struct dindex *dindex)
{
    printf ("Free entries sorted by address:\n");
    if (!avl_walk (&dindex->header->hdr_free_addr, dump_addr_entry, dindex->header))
        printf ("(none)\n");

    printf ("Free entries sorted by size and address:\n");
    if (!avl_walk (&dindex->header->hdr_free_size, dump_size_entry, dindex->header))
        printf ("(none)\n");
    return 0;
}

static int
dump_addr_entry (avl_node *freed, void *header)
{
    struct cache_free *cf = rebase (struct cache_free, cf_addr, freed);
    unsigned long beg = (char*)cf - (char*)header;
    unsigned long end = beg + cf->length - 1;
    unsigned long len = cf->length;
    return printf ("%08lx-%08lx (%08lx)\n", beg, end, len);
}

static int
dump_size_entry (avl_node *freed, void *header)
{
    struct cache_free *cf = rebase (struct cache_free, cf_size, freed);
    unsigned long beg = (char*)cf - (char*)header;
    unsigned long end = beg + cf->length - 1;
    unsigned long len = cf->length;
    return printf ("%08lx-%08lx (%08lx)\n", beg, end, len);
}
