/******************************************************************************\
 *                        ____            _ _ 
 *                       |  _ \ ___ _ __ | (_) ___ __ _
 *                       | |_) / _ \ '_ \| | |/ __/ _` |
 *                       |  _ <  __/ |_) | | | (_| (_| |
 *                       |_| \_\___| .__/|_|_|\___\__,_|
 *                                 |_|
 ******************************************************************************
 *                                 
 *  Copyright (c) 2005-2006, Herbert Nachtnebel
 *
 *    based on code from the GIT project at your favorite kernel mirror [1]
 *
 *  Copyright (c) 2005-2006, Linus Torvalds
 *                           Junio C Hamano
 *                           et al.
 *
 *  [1] http://www.kernel.org/pub/software/scm
 *  
 ******************************************************************************
 *  
 *  Title:    Replica - File Replication and Backup Utility
 *
 *  Version:  0.1
 *
 *  Date:     01-Nov-2005
 *
 *  Author:   Herbert Nachtnebel (Berri)
 *  
 ******************************************************************************
 *  
 *  Change log:
 *  
 *     Date     Rev  Ver                    Description
 *  ----------- --- ----- ---------------------------------------------------
 *  01-Nov-2005 Ber  0.1  Blatantly based on git/{cache,index}.[hc]
 *  
 ******************************************************************************
 *  
 *  Description:
 *
 *    The Directory Cache aka Index File
 *    ----------------------------------
 *
 *    The index file is a temporary cache of the current state of a directory
 *    tree.  Basically it is a database sorted by the name of the objects and
 *    holds the current metadata as well as references to the actual data of
 *    each file it knows about.
 *
 *    The most important aspect of the directory cache is that it contains a
 *    snapshot in time of the tree it caches.  Even so you can think of the
 *    cache that it represents a tree at a specific point in time, in reality
 *    the items itself are taken in a time period after each other since some
 *    processing time is needed to read the objects - just like in a backup
 *    utility.
 *
 *    The next important aspect of the directory cache is that it allows a
 *    nice split of the work needed to synchronize the directory tree with
 *    other sides - basically into two sub-tasks.  First, taking the snapshot
 *    of the directory tree which meanwhile may be used as well as altered
 *    concurrently and second, synchronizing the constant data in the cache
 *    with the other sides.  Hence, it isn't necessary to hold the mutation
 *    of the files in the directory tree at any point in time - even if that
 *    were possible.
 *
 *    The index file consists of a header and cache entries for each object
 *    in the directory tree.  For each directory an analog avl tree is held
 *    in the index file.  These avl trees are implemented in a special way.
 *    The tree node pointers are realised as offsets from there position in
 *    the file.  Therefore, wherever they are mapped into the address space
 *    of the program, no adjustments of the pointers are necessary.  Reading
 *    them in or writing them out can be realised as a simple mmap, read or
 *    write system calls, esentially making the algorithmic complexity of
 *    these operations O(1).
 *    
\******************************************************************************/
#ifndef CACHE_H_INCLUDED
#define CACHE_H_INCLUDED

/*
 *  Forward declare structure to avoid include hell.
 */
struct dindex;

/*
 *  The file signatures.
 */
#define CACHE_SIGNATURE_LE  0x4C504552  /* REPL on little endian machines */
#define CACHE_SIGNATURE_BE  0x5245504C  /* REPL on big endian machines */

#define HEADER_VERSION      0x00010001

/*
 *  An AVL tree is built from "pointers" == offsets of this type.
 *  
 *  All node "pointers" in the AVL tree are in reality offsets tagged with
 *  the skew of the node they point to.
 */
typedef int32_t avl_root;
typedef struct avl_node avl_node;

/*
 *  The AVL tree node structure.
 */
struct avl_node
{
    avl_root left;                      /* local root of left subtree */
    avl_root right;                     /* local root of right subtree */
};

/*
 *  The directory cache.
 *
 *  Note!  The fields are sorted so that 32-bit integers are stored on
 *         4 byte boundaries and 64-bit integers are stored on 8 byte
 *         boundaries.  Hence, no padding is used for the structures,
 *         both on 32-bit and 64-bit machines.
 *
 *         All fields are held in host byte order!
 */
struct cache_header 
{
    uint32_t hdr_signature;             /* signature, see above */
    uint32_t hdr_version;               /* header version */

    uint32_t hdr_dsetlen;               /* dataset length of one entry */    
    uint32_t hdr_dsetend;               /* offset behind last dataset entry */
    uint32_t hdr_filesize;              /* extend file if dsetend >= filesize */

    /*
     *  This is the root for all cache_entry structures, see below.
     */
    avl_root hdr_root;                  /* the root of all avl trees */

    /*
     *  These roots are used for memory management.
     */
    avl_root hdr_free_addr;             /* freed entries sorted by addr */
    avl_root hdr_free_size;             /* entries sorted by size and addr */

    /*
     *  I want that the cache header and cache entries have the same size.
     *  Nice to have, but in reality no algorithm enforces this.
     */
    char reserved [62];
};

/*
 *  A cache entry records the state of a file.
 *
 *  Note!  The fields are sorted so that 32-bit integers are stored on
 *         4 byte boundaries and 64-bit integers are stored on 8 byte
 *         boundaries.  Hence, no padding is used for the structures,
 *         both on 32-bit and 64-bit machines.
 *  
 *         All fields are held in host byte order.
 *
 *         If the name field gets too short, then the following cache_entry
 *         structure(s) is (are) completely used to lengthen the name field.
 */
struct cache_entry 
{
    avl_node ce_neighbors;              /* neighbor elements of a directory */
    avl_root ce_childs;                 /* avl tree root for subdir entries */
    
    uint32_t ce_mode;                   /* protection mode, file type */
    uint32_t ce_uid;                    /* owner user id */
    uint32_t ce_gid;                    /* owner group id */
    uint64_t ce_ctime;                  /* status change time (nsec's) */
    uint64_t ce_mtime;                  /* modification time  (nsec's) */
    uint32_t ce_dev;                    /* device number */
    uint32_t ce_ino;                    /* inode number */
union {
    uint32_t ce_rdev;                   /* special file, represented device */
    uint64_t ce_size;                   /* regular file, file size */
};
    uint8_t  ce_sha1 [20];              /* SHA1 hash of data */
    char     ce_name [20];              /* name of entry, variable length!!! */
};

#define S_POSIX_FIELDS 00000177777      /* POSIX bits stored in ce_mode */
#define S_MISSING_FILE 00001000000      /* tried to handle but it is missing */
#define S_BROKEN_DATA  00002000000      /* data pointed to by SHA1 is wrong */


/******************************************************************************/
/*                         Tree Add and Remove Operations                     */
/******************************************************************************/
/*
 *  Link an existing cache entry into the avl tree.
 */
struct cache_entry *link_entry (avl_root *root, struct cache_entry *key);

/*
 *  Unlink an existing cache entry from the avl tree.
 */
int unlink_entry (avl_root *root, struct cache_entry *key);


/******************************************************************************/
/*                             Tree Walking Functions                         */
/******************************************************************************/
/*
 *  Find the first cache entry in the given avl tree.
 */
struct cache_entry *find_first_entry (avl_root *root);

/*
 *  Given a cache entry, find the successor of it in the avl tree.
 */
struct cache_entry *find_next_entry (avl_root *root, struct cache_entry *key);

/*
 *  Find a specific entry in the given avl tree.
 */
struct cache_entry *find_entry (avl_root *root, const char *name);

/*
 *  Call the walker function for each entry in the avl tree.  If the walker
 *  function returns a negative value, then the operation is immediately
 *  aborted.  Else the sum of all values returned by the walker functions is
 *  returned.
 */
int walk_cache_entries (avl_root *root,
                        int (*walker)(struct cache_entry*, void*), void *data);


/******************************************************************************/
/*                             Tree Memory Management                         */
/******************************************************************************/
/*
 *  Search for a free cache entry which is in close vicinity
 *  to hint and has at least the given size.
 */
struct cache_entry *locate_free_entry (struct cache_header *hdr,
                                       avl_root *hint, size_t);

/*
 *  Remove an entry from the free lists/trees.
 */
void pull_free_entry (struct cache_header *hdr, struct cache_entry *new, size_t);

/*
 *  Append an entry to the free lists/trees.
 */
void push_free_entry (struct cache_header *hdr, struct cache_entry *old, size_t);

/******************************************************************************/

/*
 *  Dump freed objects.
 */
int dump_freed (struct dindex *dindex);

#endif /* CACHE_H_INCLUDED */

