/******************************************************************************\
 *                        ____            _ _ 
 *                       |  _ \ ___ _ __ | (_) ___ __ _
 *                       | |_) / _ \ '_ \| | |/ __/ _` |
 *                       |  _ <  __/ |_) | | | (_| (_| |
 *                       |_| \_\___| .__/|_|_|\___\__,_|
 *                                 |_|
 ******************************************************************************
 *                                 
 *  Copyright (c) 2005-2006, Herbert Nachtnebel
 *
 *    based on code from the GIT project at your favorite kernel mirror [1]
 *
 *  Copyright (c) 2005-2006, Linus Torvalds
 *                           Junio C Hamano
 *                           et al.
 *
 *  [1] http://www.kernel.org/pub/software/scm
 *  
 ******************************************************************************
 *  
 *  Title:    Replica - File Replication and Backup Utility
 *
 *  Version:  0.1
 *
 *  Date:     01-Nov-2005
 *
 *  Author:   Herbert Nachtnebel (Berri)
 *  
 ******************************************************************************
 *  
 *  Change log:
 *  
 *     Date     Rev  Ver                    Description
 *  ----------- --- ----- ---------------------------------------------------
 *  01-Nov-2005 Ber  0.1  Blatantly based on git/{cache,index}.[hc]
 *  
 ******************************************************************************
 *  
 *  Description:
 *
 *    Index File Operations
 *    ---------------------
 *
 *    In this translation unit some operations on the directory cache - the
 *    index file - are implementated.  The index file is a temporary cache
 *    of the current state of a directory tree.  Basically it is a database
 *    sorted by the name of the objects and holds the current metadata as
 *    well as references to the actual data of each file it knows about.
 *
 *    Essentially, the functions contained in here may be divided into three
 *    categories:
 *
 *      1) Low level cache [1] and index file [2] operations.
 *      2) Index file memory management.
 *      3) Index file update algorithm.
 *
 *    [1] Cache files are all files which are temporary placed in the .repl
 *        directory during update of the index file and the objects database.
 *
 *    [2] The index file holds a snapshot of the state of all files contained
 *        in a replica database.
 *    
\******************************************************************************/
#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <signal.h>

#include "compat.h"
#include "config.h"
#include "cache.h"
#include "index.h"
#include "utils.h"
#include "token.h"
#include "update.h"


static const char suffix[] = ".lock";
static const char objects[] = "/objects/XX/XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
static const char temporary[] = "/obj_XXXXXX";

static const size_t min_namelen = arrsize(((struct cache_entry*)NULL)->ce_name);
static const size_t max_double  = 128UL*1024*1024;

static struct fcache *cache_files = NULL;
static size_t page_size = 0;


/******************************************************************************/
/*                        Index File and Lock Handling                        */
/******************************************************************************/
/*
 *  remove_lock_files
 *  -----------------
 *  
 *    Unlink all registered cache (lock) files.
 */
static void
remove_lock_files (void)
{
    while (cache_files)
    {
        if (cache_files->path)
            unlink (cache_files->path);
        cache_files = cache_files->next;
    }
}

/*
 *  remove_lock_files_on_signal
 *  ---------------------------
 *  
 *    SIGINT Handler, remove all temporary files, i.e. the lock files.
 */
static void
remove_lock_files_on_signal (int signo)
{
    remove_lock_files();
    exit(1);
}

/*
 *  safe_create_leading_directories
 *  -------------------------------
 *  
 *    Create missing directories of a path.
 *
 *  Note!  We have to be extra careful not to break the system security
 *         by exposing details about the contents of protected files.
 *         The first, not really tight, security measure is to create
 *         directories in the object database with mode 0750.
 */
static int
safe_create_leading_directories (char *path)
{
    char *pos = path;

    for (pos = path; pos; *pos++ = '/')
    {
        pos = strchr (pos, '/');
        if (!pos)
            break;
        if (pos != path)
        {
            *pos = 0;
            if (mkdir (path, 0750) < 0 && errno != EEXIST)
            {
                *pos++ = '/';
                return -1;
            }    
        }
    }
    return 0;
}

/*
 *  create_temporary_cache_file
 *  ---------------------------
 *  
 *    Create a temporary cache file and register a SIGINT handler to remove
 *    it in case we are interrupted.
 *
 *  Parameters:
 *
 *    path        The directory where a temporary file should be created.
 *                To this path, a temporary file template is appended and
 *                the result is used to create the temporary file.
 *                Therefore, the path should not contain a trailing slash.
 *    
 *  Return values:
 *
 *    non-null    Pointer to a cache file control structure containing the
 *                name of the created temporary file and the opened file
 *                descriptor.
 *
 *    null        An error occured creating the temporary file, check errno
 *                for the cause.
 *  
 *  Note!  We have to be extra careful not to break the system security
 *         by exposing details about the contents of protected files.
 *         The second security measure beside the first one (see above)
 *         is handled by mkstemp for us.  It creates files with mode 0600.
 */
struct fcache *
create_temporary_cache_file (const char *path)
{
    struct fcache *cf = malloc (sizeof *cf);

    if (!cf)
        goto error;

    cf->path = malloc (strlen (path) + sizeof (temporary) + 1);
    if (!cf->path)
        goto error_drop_cache;
    sprintf (cf->path, "%s%s", path, temporary);

    if (cache_files == NULL)
    {
        signal (SIGINT, remove_lock_files_on_signal);
        atexit (remove_lock_files);
    }
    cf->next = cache_files;
    cache_files = cf;

    if (safe_create_leading_directories (cf->path) < 0)
        goto error_drop_all;
    cf->fd = mkstemp (cf->path);
    if (cf->fd < 0)
        goto error_drop_all;
    return cf;

  error_drop_all:
    cache_files = cf->next;
    free (cf->path);
    
  error_drop_cache:
    free (cf);

  error:
    return NULL;
}

/*
 *  hold_index_file_for_update
 *  --------------------------
 *
 *    Create a cache (lock) file for our index and register a SIGINT handler
 *    to remove it in case we are interrupted.
 *    
 *  Parameters:
 *
 *    path        Path to, and name of the index file.  To this path, .lock
 *                is appended to denote the lock file.  It is then tried to 
 *                create this file and open it for reading and writing.  If
 *                it already exists, then an error is returned and errno is
 *                set to EEXIST.
 *    
 *  Return values:
 *
 *    non-null    Pointer to a cache file control structure containing the
 *                name of the created index file and the opened file
 *                descriptor.
 *
 *    null        An error occured creating the temporary file, check errno
 *                for the cause.
 *
 *  TODO!  The linux man page of open states, that using O_EXCL is broken on
 *         NFS file systems.  This is ignored for now but should be handled!
 *  
 *  Note!  We have to be extra careful not to break the system security
 *         by exposing details about the contents of protected files.
 *         The index file is therefore created with protection mode 0600.
 */
struct fcache *
hold_index_file_for_update (const char *path)
{
    struct fcache *cf = malloc (sizeof *cf);

    if (!cf)
        goto error;

    cf->path = malloc (strlen (path) + sizeof (suffix) + 1);
    if (!cf->path)
        goto error_drop_cache;
    sprintf (cf->path, "%s%s", path, suffix);

    if (cache_files == NULL)
    {
        signal (SIGINT, remove_lock_files_on_signal);
        atexit (remove_lock_files);
    }
    cf->next = cache_files;
    cache_files = cf;

    if (safe_create_leading_directories (cf->path) < 0)
        goto error_drop_all;
    cf->fd = open (cf->path, O_RDWR | O_CREAT | O_EXCL, 0600);
    if (cf->fd < 0)
        goto error_drop_all;
    return cf;

  error_drop_all:
    cache_files = cf->next;
    free (cf->path);
    
  error_drop_cache:
    free (cf);

  error:
    return NULL;
}

/*
 *  open_cache_file_for_reading
 *  ---------------------------
 *  
 *    Open a cache file from our database for reading.
 *
 *  Parameters:
 *
 *    dbase       Path to directory with our database where all objects are
 *                stored.
 *
 *    sha1        The (binary) SHA1 of the object which should be opened.
 *
 *  Return values:
 *
 *    non-neg     File descriptor of opened object file.
 *
 *    -1          An error occured opening the object, check errno for cause.
 */
int
open_cache_file_for_reading (const char *dbase, const uint8_t *sha1)
{
    size_t n = strlen (dbase);
    char cache_file [n + sizeof (objects) + 1];

    memcpy (cache_file, dbase, n);
    memcpy (cache_file + n, objects, sizeof (objects) + 1);
    fill_sha1_path (cache_file + n, sha1);
    return open (cache_file, O_RDONLY);
}

/*
 *  drop_fcache
 *  -----------
 *
 *    Close a (registered) cache file and remove it from registration list.
 *
 *  Parameters:
 *
 *    cf          Pointer to cache file control structure.
 *
 *  FIXME!  It is an error not to check the return value of close!
 */
static void
drop_fcache (struct fcache *cf)
{
    struct fcache **p = &cache_files;

    close (cf->fd);
    while (*p)
    {
        if (*p == cf)
        {
            *p = cf->next;
            free (cf->path);
            free (cf);
            return;
        }
        p = &(*p)->next;
    }
    cf->path = NULL;
    cf->fd = -1;
}

/*
 *  commit_temporary_cache_file
 *  ---------------------------
 *
 *    Commit a temporary cache file to its final position given by the SHA1.
 *
 *  Parameters:
 *
 *    cf          Pointer to cache file control structure of file going to
 *                be committed.
 *
 *    sha1        The (binary) SHA1 of the object where it should be stored.
 *
 *  Return values:
 *
 *    zero        All operations worked fine, the cache file is closed now
 *                and stored at its final destination.
 *
 *    -1          An error occured, check the errno for the cause.
 *
 *  TODO!  Use link instead of rename and introduce a collision check.
 *
 *  TODO!  Check close return value.
 */
int
commit_temporary_cache_file (struct fcache *cf, const uint8_t *sha1)
{
    int rv = 0;

    if (cf->path)
    {
        int n = strlen (cf->path) - sizeof (temporary) + 1;
        char cache_file [n + sizeof (objects) + 1];
    
        memcpy (cache_file, cf->path, n);
        memcpy (cache_file + n, objects, sizeof (objects) + 1);
        fill_sha1_path (cache_file + n, sha1);
        safe_create_leading_directories (cache_file);
        rv = rename (cf->path, cache_file);
    }
    drop_fcache (cf);
    return rv;
}

/*
 *  commit_index_file
 *  -----------------
 *
 *    Update original index file with new content held in the temporary
 *    (lock) file.
 *
 *  Parameters:
 *    
 *    cf          Pointer to cache file control structure.
 *
 *  Return values:
 *
 *    zero        All operations worked fine, the index lock file is closed
 *                now and has replaced the original lock file.
 *
 *    -1          An error occured, check the errno for the cause.
 *
 *  TODO!  Check close return value.
 */
int
commit_index_file (struct fcache *cf)
{
    int rv = 0;

    if (cf->path)
    {
        int n = strlen (cf->path) - sizeof (suffix) + 1;
        char index_file [n + 1];
    
        memcpy (index_file, cf->path, n);
        index_file[n] = 0;
        rv = rename (cf->path, index_file);
    }
    drop_fcache (cf);
    return rv;
}

/*
 *  drop_cache_file
 *  ---------------
 *  
 *    Drop any updates to the index file or to cache files by closing it.
 *
 *  Parameters:
 *
 *    cf          Pointer to cache file control structure.
 *
 *  Return values:
 *
 *    zero        All went find, the file is closed now.
 *
 *    -1          An error occured, check the errno for the cause.
 */
int
drop_cache_file (struct fcache *cf)
{
    int rv = 0;
    
    if (cf->path)
        rv = unlink (cf->path);
    drop_fcache (cf);
    return rv;
}

/*
 *  verify_header
 *  -------------
 *
 *    Verify the index file header, possibly converting the index file format.
 *
 *  Parameters:
 *
 *    header      Pointer to the file header of the index file.
 *
 *    length      Length of memory mapped index file.
 *  
 *  Return values:
 *
 *    true        Index file is ok.
 *
 *    false       Errors in index file.
 *
 *  TODO!  Write the check!  Not only should the version be checked, also
 *         the endianness could be converted and all data trees should be
 *         verified.  If this is neglected, arbitrary overwrite inside our
 *         memory space might occur and in consequence, arbitrary code
 *         might be executed!
 *         
 */
int
verify_header (struct cache_header *header, size_t length)
{
    return 1;
}

/******************************************************************************/

/*
 *  create_index
 *  ------------
 *
 *    Helper function which creates the initial header for a new index file.
 */
static struct dindex *
create_index (struct dindex *di)
{
    void  *header;
    size_t length;

    if (page_size == 0)
        page_size = getpagesize();

    errno = ENOMEM;
    if (!di)
        di = malloc (sizeof *di);
    if (!di)
        return di;
    memset (di, 0, sizeof *di);

    length = (sizeof (struct cache_header) + page_size - 1) -
             (sizeof (struct cache_header) + page_size - 1) % page_size;
    header = mmap (NULL, length, PROT_READ | PROT_WRITE,
                               MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
    if (header == MAP_FAILED)
        die ("Anonymous mapping for index file failed (%s)", strerror(errno));

    di->header  = header;
    di->length  = length;
    di->changed = 0;

    di->header->hdr_signature = CACHE_SIGNATURE;
    di->header->hdr_version   = HEADER_VERSION;

    di->header->hdr_dsetlen   = sizeof (struct cache_entry);
    di->header->hdr_dsetend   = sizeof (struct cache_header);
    di->header->hdr_filesize  = length;

    di->header->hdr_root = 0;
    di->header->hdr_free_addr = 0;
    di->header->hdr_free_size = 0;
    return di;
}

/*
 *  read_index
 *  ----------
 *
 *    Read in the index file.  If there is none, create an empty index.
 *    It also initiates an initial check of the index file.
 *
 *  Parameters:
 *
 *    path        Path to, and name of the index file.
 *    
 *  Return values:
 *
 *    non-null    The index file was opened and a pointer to the index
 *                file control structure is returned.
 *
 *    null        The index file couldn't be opened.  Investigate errno
 *                for the cause.
 */
struct dindex *
read_index (const char *path)
{
    int fd, rv;
    struct stat st;
    struct dindex *di;

    char  *data;
    void  *header;
    size_t length, rounded, count;

    if (page_size == 0)
        page_size = getpagesize();

    errno = ENOMEM;
    di = malloc (sizeof *di);
    if (!di)
        return di;    
    memset (di, 0, sizeof *di);

    fd = open (path, O_RDONLY);
    if (fd < 0)    
    {
        if (errno != ENOENT)
            die ("Index file open failed (%s)", strerror(errno));

        /* no former index file exists, create an empty cache */
        return create_index (di);
    }
    fstat (fd, &st);

    /*
     *  I couldn't figure out how to avoid a SIGBUS when we extend a
     *  private mapping of the index file and write behind the file
     *  end (Note: ftruncate is no option, since I don't want to
     *  alter the original file; it is only opened for reading!).
     *
     *  Amazingly, running in a debugger and dumping the memory before
     *  writing past the end solves this problem.  Therefore, there has
     *  to be a way since the debugger is able to map the COW zero page.
     *  I had the feeling that ignoring the signal might be the solution
     *  but regretable, it did not work out.  Reading the kernel source
     *  showed the cause, only when a process is ptraced the kernel maps
     *  the zero page for non existing pages past the file end :-(
     *
     *  I am going to solve this in a portable way.  Instead of mapping 
     *  the index file I read it into a private mapping large enough to 
     *  hold it.
     *
     *  Historical Note: in ancient linux times the kernel mapped the
     *  COW zero page for private, _read-only_ mappings past the file
     *  end.  Alas, these times are gone now.  But why???
     */
    length  = (st.st_size + page_size - 1) -
              (st.st_size + page_size - 1) % page_size;
    rounded = (st.st_size + max_double) - (st.st_size + max_double) % max_double;

    /*
     *  1st, place a fairly large mapping into our memory space.
     *
     *  Note!  We make the mapping immediately smaller but with this,
     *         later extensions up to the size allocated here will not
     *         move the mapping in our memory space!
     */
    header = mmap (NULL, rounded, PROT_READ | PROT_WRITE,
                   MAP_PRIVATE | MAP_NORESERVE | MAP_ANONYMOUS, 0, 0);
    if (header == MAP_FAILED)
        die ("Anonymous mapping for index file failed (%s)", strerror(errno));
    di->header = header;
    di->length = rounded;

    /*
     *  2nd, read-in the index file.
     */
    count = st.st_size;
    data  = header;
    while (count)
    {
        ssize_t got = read (fd, data, count);

        if (got == -1)
        {
            if (errno != EINTR)
                goto error;
        }
        else if (got == 0)
        {
            errno = EIO;
            goto error;
        }
        else
        {
            data  += got;
            count -= got;
        }
    }

    /*
     *  3rd, now make the mapping reasonably long.
     */
    close (fd);
    header = mremap (header, rounded, length, 0);
    di->header = header;
    di->length = length;

    errno = EINVAL;
    if (verify_header (header, length) < 0)
        goto error;    

    return di;

  error:
    rv = errno;
    munmap (di->header, di->length);
    free (di);
    errno = rv;
    return NULL;
}

/*
 *  peek_at_index
 *  -------------
 *  
 *    This function is similar to read_index, but makes a read-only mapping
 *    of the index file.  To be correct, writing to the private mapping is
 *    not prohibited, but the mapping itself can _not_ be extended.  Hence,
 *    operations which modify the index file in place _could_ use this but
 *    nevertheless, they should use read_index instead.
 *
 *  Parameters:
 *
 *    path        Path to, and name of the index file.
 *    
 *  Return values:
 *
 *    non-null    The index file was opened and a pointer to the index
 *                file control structure is returned.
 *
 *    null        The index file couldn't be opened.  Investigate errno
 *                for the cause.
 *
 *  Note!  Many operations only require a read-only access to the
 *         index file.  Since the update algorithm of the index file
 *         never touches the original index file and atomically
 *         replaces the existing index file with an update, it is
 *         always safe to take a read-only look at the index file.
 *         Consequently, read-only operations on the index never
 *         have to take the index file lock.
 */
struct dindex *
peek_at_index (const char *path)
{
    int fd, rv;
    struct stat st;
    struct dindex *di;

    void  *header;
    size_t length;

    if (page_size == 0)
        page_size = getpagesize();

    errno = ENOMEM;
    di = malloc (sizeof *di);
    if (!di)
        return di;    
    memset (di, 0, sizeof *di);

    fd = open (path, O_RDONLY);
    if (fd < 0)    
    {
        if (errno != ENOENT)
            die ("Index file open failed (%s)", strerror(errno));
        return NULL;
    }
    fstat (fd, &st);

    length = (st.st_size + page_size - 1) -
             (st.st_size + page_size - 1) % page_size;
    header = mmap (NULL, length, PROT_READ | PROT_WRITE, MAP_PRIVATE, fd, 0);
    if (header == MAP_FAILED)
        die ("Index file mapping failed (%s)", strerror(errno));
    close (fd);

    di->header  = header;
    di->length  = length;
    di->changed = 0;

    errno = EINVAL;
    if (verify_header (header, length) < 0)
        goto error;
    return di;

  error:
    rv = errno;
    munmap (di->header, di->length);
    free (di);
    errno = rv;
    return NULL;
}

/*
 *  write_index
 *  -----------
 *
 *    Write the current internal database into the index lock file.
 *
 *  Parameters:
 *
 *    cf          Pointer to control structure of index lock file.
 *
 *    dindex      Pointer to header of index file.
 *
 *  Return values:
 *
 *    zero        All went find, the file is updated now.
 *
 *    -1          An error occured, check the errno for the cause.
 *
 *  Note!  It's so really simple and fast, look...
 */
int
write_index (struct fcache *cf, struct dindex *dindex)
{
    ssize_t count;
    size_t put = dindex->header->hdr_dsetend;
    char *data = (char*) dindex->header;

    dindex->header->hdr_filesize = put;
    while (put)
    {
        count = write (cf->fd, data, put);
        if (count == -1)
        {
            if (errno != EINTR)
                return -1;
        }
        else
        {
            data += count;
            put  -= count;
        }
    }
    return 0;
}


/******************************************************************************/
/*                        Index File Memory Management                        */
/******************************************************************************/
/*
 *  reserve_entry
 *  -------------
 *
 *    Look for an empty cache_entry in our memory map of the index file
 *    large enough to hold the entry with the given name.
 *
 *  Parameters:
 *
 *    dindex       Pointer to the memory mapped index file header.
 *
 *    hint         The new object should be allocated near to this root.
 *
 *    name         The name of the newly created cache file entry.
 *
 *  Return values:
 *
 *    non-null     The pointer to the cache file entry reserved for the
 *                 object.  The name field of the entry was filled.
 *
 *    null         There were not enough memory the allocate the entry.
 *                 The errno is set for a more specific error description.
 *    
 *  Properties:
 *
 *    1st) When we reuse a previously allocated entry, we depend on the
 *         fact that it was previously cleared completely.
 *
 *    2nd) If we have to extend the file mapping, the complete file may be
 *         moved freely in our memory space.  All callers have to readjust
 *         their stored pointers into this memory space!
 *
 *    3rd) We only return a pointer to a temporary cache_entry which may be
 *         allocated subsequently.  We do _not_ take this entry from any
 *         free list!
 *
 *    4th) Hence, an entry can be bound concurrently to both, the actual
 *         data tree, and to the free list.  That may sound strange, but it
 *         simplifies the logic and allows for a nice split of the memory
 *         management and the low level avl tree implementation.
 */
static struct cache_entry *
reserve_entry (struct dindex *dindex, avl_root *hint, const char *name)
{
    struct cache_entry *rv;
    size_t namelen = strlen (name);
    size_t alloced = dindex->header->hdr_dsetlen;

    if (namelen >= min_namelen)
        alloced += namelen - min_namelen  + alloced -
                  (namelen - min_namelen) % alloced;

    /*
     *  Search for a free cache_entry structure large enough to hold the given
     *  name which is in close vicinity to the tree root.
     */
    rv = locate_free_entry (dindex->header, hint, alloced);
    if (rv)
        goto set_name;

    /*
     *  No suitable--previously freed--cache entry, append a new one.
     */
    if (dindex->header->hdr_dsetend + alloced > dindex->length)
    {
        size_t len;
        void  *map;
        
        if (page_size == 0)
            page_size = getpagesize();

        /*
         *  We double the used memory space until it gets larger than a
         *  large constant where we switch to linear expansion.
         */
        if (dindex->length > max_double)
            len = (dindex->length + max_double) -
                  (dindex->length + max_double) % max_double;
        else
            len = 2*dindex->length;

        /*
         *  Extend the existing memory map.
         */
        map = mremap (dindex->header, dindex->length, len, MREMAP_MAYMOVE);
        if (map == MAP_FAILED)
            return NULL;
        dindex->header = map;
        dindex->length = len;
    }
    rv = (void*)dindex->header + dindex->header->hdr_dsetend;

  set_name:
    memcpy (rv->ce_name, name, namelen+1);
    return rv;
}

/*
 *  finalize_entry
 *  --------------
 *  
 *    Finalize the allocation of a new cache entry.
 *
 *    The object has to come from a call to the function reserve_entry above
 *    and is therefore considered to be only partially constructed.  The entry
 *    is joined to the tree and pulled from the free list.  Hence, after that,
 *    the new object is a valid data object and in the next step the metadata
 *    may be filled in.
 *
 *  Parameters:
 *
 *    dindex       Pointer to the memory mapped index file header.
 *
 *    obj          Pointer to a cache entry got previously from reserver_entry.
 *
 *  Return values:
 *
 *    non-null     The passed in object pointer obj is returned unchanged.
 */
static struct cache_entry*
finalize_entry (struct dindex *dindex, struct cache_entry *obj)
{
    size_t namelen = strlen (obj->ce_name);
    size_t alloced = dindex->header->hdr_dsetlen;
    char *beg, *end;

    if (namelen >= min_namelen)
        alloced += namelen - min_namelen  + alloced -
                  (namelen - min_namelen) % alloced;

    /*
     *  Whipe out the unused part of the ce_name field.
     */
    beg = obj->ce_name + namelen + 1;
    end = (char*)obj + alloced;
    memset (beg, 0, end-beg);

    /*
     *  Remove entry from free lists.
     */
    if (obj == (void*)dindex->header + dindex->header->hdr_dsetend)
        dindex->header->hdr_dsetend += alloced;
    else
        pull_free_entry (dindex->header, obj, alloced);

    return obj;
}

/*
 *  adjust_position
 *  ---------------
 *
 *    Readjust a pointer if the memory mapped index file was moved in our    
 *    address space.
 */
static inline void*
adjust_position (void *adjusted, void *oldpos, void *newpos)
{
    return (char*)adjusted + ((char*)newpos - (char*)oldpos);
}

/*
 *  delete_cache_entries
 *  --------------------
 *
 *    Delete a cache entry as well as all of its subentries.
 *
 *  Parameters:
 *
 *    dindex       Pointer to the memory mapped index file header.
 *
 *    tree         Pointer to the tree root where the object is deleted from.
 *
 *    ce           Pointer to a cache entry which is going to be deleted.
 *                 If this entry is a root for other entries, then all
 *                 subentries are delete as well.
 *
 *  Return values:
 *
 *    true        The object could be deleted.
 *
 *    false       The object could not be found.
 */
static int
delete_cache_entries (struct dindex *dindex,
                      avl_root *tree, struct cache_entry *ce)
{
    /*
     *  Unlink this entry from its tree.
     */
    int rv = unlink_entry (tree, ce);

    if (rv)
    {
        size_t namelen = strlen (ce->ce_name);
        size_t alloced = dindex->header->hdr_dsetlen;
        
        avl_root *subtree = &ce->ce_childs;
        struct cache_entry *child = find_first_entry (subtree);

        if (namelen >= min_namelen)
            alloced += namelen - min_namelen  + alloced -
                      (namelen - min_namelen) % alloced;

        /*
         *  Delete all subentries.
         */
        while (child)
        {
            struct cache_entry *next = find_next_entry (subtree, child);

            rv += delete_cache_entries (dindex, subtree, child);
            child = next;
        }

        /*
         *  Finally, this entry is pushed to the free "list".
         */
        push_free_entry (dindex->header, ce, alloced);
    }
    return rv;
}


/******************************************************************************/
/*                         Index File Update Algorithm                        */
/******************************************************************************/
/*
 *  verify_filename
 *  ---------------
 *
 *    Check whether the filename given should be included into the directory
 *    index.  Some names are explicitely forbidden (i.e. ".", ".."), others
 *    can be excluded via a user supplied filter.
 *
 *  Parameters:
 *
 *    filter      Regular expression used to filter out filenames.
 *
 *    path        Directory where name is a member.
 *
 *    name        Filename to check.
 *
 *  Return values:
 *
 *    true        The file should be added to the directory index.
 *
 *    false       The file should be excluded from the index.
 *
 *  TODO!  Implement it!
 */
static int
verify_filename (struct regexpr *filter, const char *path, const char *name)
{
    if (strcmp (name, ".")  == 0 ||
        strcmp (name, "..") == 0 ||
        strcmp (name, ".repl") == 0)
        return 0;
    return 1;
}

/*
 *  is_missing
 *  ----------
 *
 *    Given a cache entry, check whether the described file exists in the
 *    file system.
 *
 *  Parameters:
 *
 *    ce          Pointer to the cache entry.
 *
 *    path        Directory where the file resides.
 *
 *    nocd        Less then zero if the path should be joined with the
 *                filename, zero if we previously changed into that
 *                directory and the name of the file is sufficient alone.
 *
 *  Return values:
 *
 *    true        The object exists in the file system.
 *
 *    false       No matching file system entry could be found.
 */
static int
is_missing (struct cache_entry *ce, const char *path, int nocd)
{
    if (ce)
    {
        struct stat st;
        size_t path_len = strlen (path);
        size_t name_len = strlen (ce->ce_name);
        char file [path_len + name_len + 2];

        if (nocd < 0)
        {
            memcpy (file, path, path_len + 1);
            memcpy (file + path_len + 1, ce->ce_name, name_len + 1);
            file [path_len] = '/';
        }
        else
        {
            memcpy (file, ce->ce_name, name_len + 1);
        }
        if (lstat (file, &st) < 0 && errno == ENOENT)
            return 1;
    }
    return 0;
}

/*
 *  is_directory
 *  ------------
 *
 *    Checks whether the cache entry represents a directory or not.
 */
static int
is_directory (struct cache_entry *ce)
{
    if (ce)
    {
        if (S_ISDIR(ce->ce_mode))
            return 1;
    }
    return 0;
}

/*
 *  remove_one
 *  ----------
 *
 *    Remove a cache entry from the index.  If the cache entry represents
 *    a directory in the file system, then all subentries are also deleted.
 *
 *  Parameters:
 *
 *    dindex      Pointer to the memory mapped index file used now.
 *
 *    tree        Pointer to the root of the tree from where the object
 *                is deleted.
 *
 *    ce          Pointer to a cache entry which is going to be deleted.
 *
 *  Return values:
 *
 *    true        The object could be deleted.
 *
 *    false       The object could not be found.
 */
static int
remove_one (struct dindex *dindex, avl_root *tree, struct cache_entry *ce)
{
    if (delete_cache_entries (dindex, tree, ce))
        return dindex->changed = 1;
    return 0;
}

/*
 *  update_one
 *  ----------
 *
 *    Checks whether our internal cache has to be modified for a single file
 *    system element and updates it as well.  If new entries are found, it
 *    joins them to our index.  Updated content of the file are read-in and
 *    placed into our object database.
 *
 *  Parameters:
 *
 *    config      Pointer to our actual configuration.
 *
 *    dindex      Pointer to the memory mapped index file.
 *
 *    tree        Pointer to the tree root where the element is a member.
 *
 *    path        Directory where the element resides in the file system.
 *
 *    name        Name of the file system element refered to.
 *
 *    nocd        Did we previously change into the given directory.
 *
 *  Return values:
 *
 *    non-null    Pointer to updated cache entry.
 *
 *    null        An error occured during the update.
 *                The errno is set according to the cause.
 */
static struct cache_entry *
update_one (struct config *config, struct dindex *dindex, avl_root *tree,
            const char *path, const char *name, int nocd)
{
    struct stat st;
    struct cache_entry *ce, *new;
    struct cache_header *header = dindex->header;
    size_t path_len = strlen (path);
    size_t name_len = strlen (name);
    char file [path_len + name_len + 2];

    /*
     *  Path to entry we operate against, three cases:
     */
    if (nocd > 0)
    {
        /*
         *  Case 1, a root directory entry, just use the path information.
         */
        memcpy (file, path, path_len + 1);
    }
    else if (nocd < 0)
    {
        /*
         *  Case 2, we need an absolute path to the entry, join it.
         */
        memcpy (file, path, path_len + 1);
        memcpy (file + path_len + 1, name, name_len + 1);
        file [path_len] = '/';
    }
    else
    {
        /*
         *  Case 3, we changed into the given directory, just use the name.
         */
        memcpy (file, name, name_len + 1);
    }    

    /*
     *  Get information about the directory entry.
     */
    if (lstat (file, &st) < 0)
        return NULL;

    /*
     *  Check the file types, all others are silently ignored.
     */
    errno = ECANCELED;
    if (!S_ISREG(st.st_mode) &&
        !S_ISDIR(st.st_mode) &&
        !S_ISCHR(st.st_mode) &&
        !S_ISBLK(st.st_mode) &&
        !S_ISLNK(st.st_mode) &&
        !S_ISFIFO(st.st_mode))
        return NULL;

    /*
     * Find a temporary cache file entry and fill-in the name.
     *
     * Note!  Reserve_entry allocates a temporary entry in the index.
     *        If this entry is a previously freed entry, then, at this
     *        point in time, this entry will still be on the free "list".
     *        Hence, only the name field of this cache entry may be used!
     */
    new = reserve_entry (dindex, tree, name);
    errno = ENOMEM;
    if (new == NULL)
        return NULL;
    if (header != dindex->header)
        tree = adjust_position (tree, header, dindex->header);

    /*
     *  Now check if we already had such a cache entry.  If not, add the
     *  temporary entry to the tree.
     *
     *  Note!  Link_entry links the temporary entry to the "life" index
     *         if it doesn't exist already.  After that, the cache entry
     *         is bound to the index and to the free list concurrently.
     *         Finalize_entry then takes the entry from the free list
     *         and makes it a freely usable new one.
     */
    ce = link_entry (tree, new);
    if (ce == new)
    {
        finalize_entry (dindex, ce);
        if (S_ISDIR(st.st_mode))
            info ("New directory %s/%s", path, name);
    }

    /*
     *  Ok, lets check the file metadata and update it as well as its contents.
     */
    switch (update_metadata (ce, &st))
    {
        case FILEDATA_CHANGED:
            if (update_filedata (ce, file, config->dbase) < 0)
            {
                /* Broken file, complain about it */
                info ("Error reading %s/%s: %s", path, name, strerror(errno));
            }
            /* fall through */

        case METADATA_CHANGED:
            dindex->changed = 1;

        default:
            break;
    }
    return ce;
}

/******************************************************************************/

/*
 *  update_one_dir
 *  --------------
 *
 *    Recursivly read-in directories and update the index as well as our
 *    object database with the file metadata and the file contents found.
 *
 *  Parameters:
 *
 *    config      Pointer to our actual configuration.
 *
 *    dindex      Pointer to the memory mapped index file.
 *
 *    path        Directory where the element resides in the file system.
 *
 *    name        Name of the file system element refered to.
 *
 *    nocd        Did we previously change into the given directory.
 *
 *  Return values:
 *
 *    zero        No errors occured reading in the directory tree.
 *
 *    -1          An error occured during the update.
 *                The errno is set according to the cause.
 */
static int
update_one_dir (struct config *config, struct dindex *dindex, avl_root *tree,
                const char *path, const char *name, int nocd)
{
    DIR *dir;
    struct dirent *de;
    struct cache_entry *p;

    /*
     *  Directory we operate against.
     */
    size_t path_len = strlen (path);
    size_t name_len = strlen (name);
    char cwd [path_len + name_len + 2];

    memcpy (cwd, path, path_len + 1);
    if (nocd < 1)
    {
        memcpy (cwd + path_len + 1, name, name_len + 1);
        cwd [path_len] = '/';
    }

    /*
     *  1st, loop over the directory contents, adding the found entries.
     */
    dir = opendir (nocd ? cwd : name);
    if (!dir)
        return -1;

    p = update_one (config, dindex, tree, path, name, nocd);
    if (!p)
    {
        int rv = errno;
        closedir (dir);
        errno = rv;
        return -1;
    }
    tree = &p->ce_childs;
    
    if (nocd)
        nocd--;
    if (nocd == 0)
        nocd = fchdir (dirfd (dir));

    while (NULL != (de = readdir (dir)))
        if (verify_filename (config->filter, cwd, de->d_name))
        {
            struct cache_header *header = dindex->header;
            
            update_one (config, dindex, tree, cwd, de->d_name, nocd);
            if (header != dindex->header)
                tree = adjust_position (tree, header, dindex->header);
        }
    closedir (dir);
    
    /*
     *  2nd, loop over cache entries, recursively updating subdirs.
     */
    for (p = find_first_entry (tree); p; p = find_next_entry (tree, p))
    {
        while (is_missing (p, cwd, nocd))
        {
            struct cache_entry *next = find_next_entry (tree, p);

            remove_one (dindex, tree, p);
            p = next;
        }

        if (is_directory (p))
        {
            struct cache_header *header = dindex->header;
            
            update_one_dir (config, dindex, tree, cwd, p->ce_name, nocd);
            if (header != dindex->header)
            {
                p = adjust_position (p, header, dindex->header);
                tree = adjust_position (tree, header, dindex->header);
            }
        }
    }

    /*
     *  Try to change the directory back to the previous one.
     */
    return (nocd == 0) ? chdir("..") : 0;
}

/*
 *  update_index
 *  ------------
 *
 *    High level index update algorithm.  Reads in all the configured
 *    root directories as well as their contents and updates the index
 *    file and the object database with the file metadata and the file
 *    contents found there.
 *
 *  Parameters:
 *
 *    config      Pointer to our actual configuration.
 *
 *    dindex      Pointer to the memory mapped index file.
 *
 *  Return values:
 *
 *    zero        No errors occured reading in the directory trees.
 *
 *    -1          An error occured during the update.
 *                The errno is set according to the cause.
 */
int
update_index (struct config *config, struct dindex *dindex)
{
    int rv;
    char **roots = config->roots;
    char **names = config->names;
    DIR *old = opendir (".");

    for (rv = 0; *roots && *names; roots++, names++)
    {
        if (update_one_dir (config, dindex,
                            &dindex->header->hdr_root,
                            *roots, *names, 1) < 0)
            rv = -1;
        fchdir (dirfd (old));
    }
    closedir (old);
    return rv;
}

/******************************************************************************/

static int
dump_cache_entry (struct cache_entry *ce, void *path)
{
    size_t path_len = path ? strlen (path) : 0;
    size_t ce_len = strlen (ce->ce_name);
    char name [path_len + ce_len + 2];
    char type, ppsha1 [41];

    /*
     *  Convert SHA1 into a string.
     */
    format_sha1_str (ppsha1, ce->ce_sha1);

    /*
     *  Combine name with the path to it.
     */
    memcpy (name, path, path_len);
    name[path_len] = '/';
    memcpy (name + path_len + 1, ce->ce_name, ce_len);
    name[path_len + ce_len + 1] = 0;

    /*
     *  What's the type of this entry?
     */
    if (S_ISREG(ce->ce_mode))
        type = 'R';
    else if (S_ISDIR(ce->ce_mode))
        type = 'D';
    else if (S_ISCHR(ce->ce_mode))
        type = 'C';
    else if (S_ISBLK(ce->ce_mode))
        type = 'B';
    else if (S_ISLNK(ce->ce_mode))
        type = 'L';
    else if (S_ISFIFO(ce->ce_mode))
        type = 'F';
    else
        type = '?';

    /*
     *  Finally we have all informations to print it.
     */
    printf ("%c %o %d:%d %08llu %s %s\n",
            type, ce->ce_mode & 07777,
            ce->ce_uid, ce->ce_gid, ce->ce_size, ppsha1, name);

    /*
     *  If this entry has subentries, walk them also.
     */
    if (ce->ce_childs)
        return walk_cache_entries (&ce->ce_childs, dump_cache_entry, name);
    return 0;
}

int
dump_index (struct dindex *dindex)
{
    return walk_cache_entries (&dindex->header->hdr_root, dump_cache_entry, NULL);
}

