/******************************************************************************\
 *                        ____            _ _ 
 *                       |  _ \ ___ _ __ | (_) ___ __ _
 *                       | |_) / _ \ '_ \| | |/ __/ _` |
 *                       |  _ <  __/ |_) | | | (_| (_| |
 *                       |_| \_\___| .__/|_|_|\___\__,_|
 *                                 |_|
 ******************************************************************************
 *                                 
 *  Copyright (c) 2005-2006, Herbert Nachtnebel
 *
 *    based on code from the GIT project at your favorite kernel mirror [1]
 *
 *  Copyright (c) 2005-2006, Linus Torvalds
 *                           Junio C Hamano
 *                           et al.
 *
 *  [1] http://www.kernel.org/pub/software/scm
 *                      
 ******************************************************************************
 *                      
 *  Title:    Replica - File Replication and Backup Utility
 *
 *  Version:  0.1
 *
 *  Date:     01-Nov-2005
 *
 *  Author:   Herbert Nachtnebel (Berri)
 *  
 ******************************************************************************
 *  
 *  Change log:
 *  
 *     Date     Rev  Ver                    Description
 *  ----------- --- ----- ---------------------------------------------------
 *  01-Nov-2005 Ber  0.1  Blatantly based on git/{cache,index}.[hc]
 *  
 ******************************************************************************
 *  
 *  Description:
 *
 *    Index and Cache File Operations
 *    -------------------------------
 *
 *    The index file is a temporary cache of the current state of a directory
 *    tree.  Basically it is a database sorted by the name of the objects and
 *    holds the current metadata as well as references to the actual data of
 *    each file it knows about.
 *
 *    Cache files are all files, which are temporary placed in the .repl
 *    directory during updating the index file as well as the object database.
 *    All cache file control structures are internally linked into a linear
 *    list used to delete opened temporary files in case the application is
 *    interrupted and terminated.
 *    
\******************************************************************************/
#ifndef INDEX_H_INCLUDED
#define INDEX_H_INCLUDED

/*
 *  Forward declare structure to avoid include hell.
 */
struct config;

/*
 *  Internal control structure for the index file.
 */
struct dindex
{
    struct cache_header *header;        /* memory map of index file */
    size_t length;                      /* length of mapped area */
    int changed;                        /* we made some changes to the file */
};

/*
 *  Internal control structure for cache files.
 */
struct fcache
{
    struct fcache *next;
    char *path;
    int fd;
};
    
/******************************************************************************/

/*
 *  Create a temporary cache file in our database.
 */
struct fcache *create_temporary_cache_file (const char *path);

/*
 *  Commit the temporary cache file to its final position given by the SHA1.
 */
int commit_temporary_cache_file (struct fcache *cf, const uint8_t *sha1);

/*
 *  Create a temporary file for locking and updating the index file.
 */
struct fcache *hold_index_file_for_update (const char *path);

/*
 *  Open a cache file from our database.
 */
int open_cache_file_for_reading (const char *dbase, const uint8_t *sha1);

/*
 *  Read the index file into the internal database.
 */
struct dindex *read_index (const char *path);

/*
 *  Make a read-only mapping of the index file.
 */
struct dindex *peek_at_index (const char *path);

/*
 *  Bring the index file into accordance with the filesystem.
 */
int update_index (struct config *config, struct dindex *dindex);

/*
 *  Write the current internal database into the temporary index file.
 */
int write_index (struct fcache *fcache, struct dindex *dindex);

/*
 *  Commit the updated index back to the original index file.
 */
int commit_index_file (struct fcache *cf);

/*
 *  Drop the cache file, together with the data (maybe updated) contained in it.
 */
int drop_cache_file (struct fcache *cf);

/******************************************************************************/

/*
 *  Dump the internal index database.
 */
int dump_index (struct dindex *dindex);

#endif /* INDEX_H_INCLUDED */
