/******************************************************************************\
 *                        ____            _ _ 
 *                       |  _ \ ___ _ __ | (_) ___ __ _
 *                       | |_) / _ \ '_ \| | |/ __/ _` |
 *                       |  _ <  __/ |_) | | | (_| (_| |
 *                       |_| \_\___| .__/|_|_|\___\__,_|
 *                                 |_|
 ******************************************************************************
 *                                 
 *  Copyright (c) 2005-2006, Herbert Nachtnebel
 *  
 ******************************************************************************
 *  
 *  Title:    Replica - File Replication and Backup Utility
 *
 *  Version:  0.1
 *
 *  Date:     01-Nov-2005
 *
 *  Author:   Herbert Nachtnebel (Berri)
 *  
 ******************************************************************************
 *  
 *  Change log:
 *  
 *     Date     Rev  Ver                    Description
 *  ----------- --- ----- ---------------------------------------------------
 *  01-Nov-2005 Ber  0.1  Initial implementation
 *  
 ******************************************************************************
 *  
 *  Description:
 *
 *    Main Program State Machine
 *    --------------------------
 *    
\******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/utsname.h>

#include "compat.h"
#include "config.h"
#include "index.h"
#include "cache.h"
#include "utils.h"
#include "token.h"
#include "update.h"
#include "tree.h"


/*
 *  The index file is a form of a directory cache. It records the
 *  state of each file in a tree at a specific point in time.
 *
 *  1.) update index file
 *  2.) write index to commit objects
 *  3.) transfer commit objects to other sides
 *  4.) merge with commit objects of other sides
 *  5.) collect changes to files (inotify)
 *  6.) continue at 2.)
 *
 */

struct config *
read_config (const char *path)
{
    static const char *roots[]   = { "./test1", "./test2", NULL };
    static const char *names[]   = {  "set-1",   "set-2",  NULL };
    static const char *exclude[] = { "*~", "*.bak", NULL };

    char cwd [16384];
    struct utsname uts;

    static struct config test_conf =
    {
        .nodename = "(none)",
        .codepage = "ISO-8859-15",
        .dbase    = "./.repl",
        .index    = "./.repl/index",
        
        .roots    = roots,
        .names    = names,
        .exclude  = exclude
    };

    if (uname (&uts) == 0)
    {
        test_conf.nodename = strdup (uts.nodename);
    }

    if (getcwd (cwd, sizeof cwd))
    {
        if (test_conf.dbase[0] == '.' && test_conf.dbase[1] == '/')
        {
            char *new = malloc (strlen (test_conf.dbase) + strlen (cwd));
            strcpy (new, cwd);
            strcat (new, test_conf.dbase+1);
            test_conf.dbase = new;
        }
        if (test_conf.index[0] == '.' && test_conf.index[1] == '/')
        {
            char *new = malloc (strlen (test_conf.index) + strlen (cwd));
            strcpy (new, cwd);
            strcat (new, test_conf.index+1);
            test_conf.index = new;
        }
    }
    return &test_conf;
}

int
main (int argc, char **argv)
{
    struct config *config;
    struct fcache *fcache;
    struct dindex *dindex;

    /*
     *  Read the current configuration from a file or from a
     *  remote database.
     */
    config = read_config (argv[1]);
    if (!config)
        die ("No configuration available");

    if (argc > 1)
    {
        dindex = peek_at_index (config->index);
        if (!dindex)
            die ("Error opening index file");
        
        if (strcasecmp (argv[1], "dump-index") == 0)
        {
            dump_index (dindex);
        }
        else if (strcasecmp (argv[1], "dump-freed") == 0)
        {
            dump_freed (dindex);
        }
        else if (strcasecmp (argv[1], "type-of") == 0)
        {
            dump_type_of (config->dbase, argv[2]);
        }
        else if (strcasecmp (argv[1], "sha1-cat") == 0)
        {
            dump_sha1_cat (config->dbase, argv[2]);
        }
        else if (strcasecmp (argv[1], "write-tree") == 0)
        {
            if (write_tree (config, dindex, NULL) < 0)
                die ("Error writing tree");
        }
        else if (strcasecmp (argv[1], "commit") == 0)
        {
            uint8_t sha1[20];
            
            if (write_tree (config, dindex, sha1) < 0)
                die ("Error writing tree");
            
            if (commit_tree (config, sha1, NULL) < 0)
                die ("Error commiting tree");
        }
    }
    else
    {
        /*
         *  Create temporary index file.  It has two purposes.
         *  First, it is a flag holding other applications back
         *  to fiddle with the cache.  Second, it serves as a
         *  container for the updated cache.
         */
        fcache = hold_index_file_for_update (config->index);
        if (!fcache)
            die ("Unable to create temporary index file");

        /*
         *  Read in the current directory cache.
         */
        dindex = read_index (config->index);
        if (!dindex)
            die ("Index file corrupted");

        /*
         *  Update the directory cache from the filesystem.
         */
        if (update_index (config, dindex) < 0)
            die ("Error updating index file");

        /*
         *  If it has changed, rewrite it.
         */
        if (dindex->changed)
        {
            if (write_index (fcache, dindex) < 0)
                die ("Error writing index file");

            if (commit_index_file (fcache) < 0)
                die ("Unable to commit index file");
        }
        else
        {
            drop_cache_file (fcache);
        }
    }
    return 0;
}
