/******************************************************************************\
 *                        ____            _ _ 
 *                       |  _ \ ___ _ __ | (_) ___ __ _
 *                       | |_) / _ \ '_ \| | |/ __/ _` |
 *                       |  _ <  __/ |_) | | | (_| (_| |
 *                       |_| \_\___| .__/|_|_|\___\__,_|
 *                                 |_|
 ******************************************************************************
 *                                 
 *  Copyright (c) 2005-2006, Herbert Nachtnebel
 *
 *    based on code from the GIT project at your favorite kernel mirror [1]
 *
 *  Copyright (c) 2005-2006, Linus Torvalds
 *                           Junio C Hamano
 *                           et al.
 *
 *  [1] http://www.kernel.org/pub/software/scm
 *  
 ******************************************************************************
 *                      
 *  Title:    Replica - File Replication and Backup Utility
 *
 *  Version:  0.1
 *
 *  Date:     01-Nov-2005
 *
 *  Author:   Herbert Nachtnebel (Berri)
 *  
 ******************************************************************************
 *  
 *  Change log:
 *  
 *     Date     Rev  Ver                    Description
 *  ----------- --- ----- ---------------------------------------------------
 *  01-Nov-2005 Ber  0.1  Initial implementation
 *  
 ******************************************************************************
 *  
 *  Description:
 *
 *    Tree Operations
 *    ---------------
 *    
\******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <memory.h>
#include <errno.h>

#include "compat.h"
#include "config.h"
#include "cache.h"
#include "index.h"
#include "tree.h"
#include "utils.h"
#include "token.h"
#include "update.h"


static int write_subdirs();
static int fill_tree_entries();

static int
write_tree_inner (struct config *config, avl_root *root, uint8_t *sha1)
{
    /*
     *  Walk through this directory writing out subdir's.
     *
     *  Note! While writing out the subdirectories, we also count
     *        how many bytes will be written to this tree object.
     */
    int size = walk_cache_entries (root, write_subdirs, config);

    if (size >= 0)
    {
        /*
         *  Create a temporary buffer large enough to hold the tree object.
         */
        char tree [size];

        /*
         *  Enter entries into the tree object.
         */
        config->private = tree;
        if (walk_cache_entries (root, fill_tree_entries, config) < 0)
            return -1;

        /*
         *  Write the result to our database.
         */
        return write_sha1_file (config->dbase, T_TREE, T_POSIX, tree, size, sha1);
    }
    return -1;
}

static int
write_subdirs (struct cache_entry *ce, struct config *config)
{
    token utf;
    size_t nam_len;
    
    /*
     *  Write out all subdirectories of a directory cache entry.
     */
    if (ce->ce_childs &&
        write_tree_inner (config, &ce->ce_childs, ce->ce_sha1) < 0)
        return -1;

    /*
     *  Estimate the length needed for the name of this entry.
     */
    nam_len = loc2utf( config->codepage, &utf,
                       ce->ce_name, strlen(ce->ce_name) + 1, NULL, 0 );

    /*
     *  Remember what's the most efficient encoding for this entry.
     */
    if (utf == T_UTF_8)
        ce->ce_mode &= ~S_UTF16BE;
    else if (utf == T_UTF_16BE)
        ce->ce_mode |= S_UTF16BE;
    else
        die ("Invalid conversion codepage: %s", utf);    

    /*
     *  Return the size needed by this entry in the tree object.
     */
    return sizeof (struct tree_entry) + nam_len;
}

static int
fill_tree_entries (struct cache_entry *ce, struct config *config)
{
    struct tree_entry align, *te = (struct tree_entry*) config->private;
    size_t nam_len;

    /*
     *  Copy the fields while adjusting the byte order.
     *
     *  Note!  This is for architectures which can't handle unaligned data!
     */
    align.te_ctime = cpu2host_64 (ce->ce_ctime);
    align.te_mtime = cpu2host_64 (ce->ce_mtime);
    align.te_mode  = cpu2host_32 (ce->ce_mode);
    align.te_uid   = cpu2host_32 (ce->ce_uid);
    align.te_gid   = cpu2host_32 (ce->ce_gid);

    /*
     *  No need to transform the sha1, just copy it over.
     */
    memcpy (align.te_sha1, ce->ce_sha1, sizeof align.te_sha1);

    /*
     *  Now move the entry to its final place.
     */
    memcpy (te, &align, sizeof align);

    /*
     *  Encode the name of this object in Unicode.
     *  
     *  Note!  The tree entry was allocated large enough, therefore we
     *         can avoid to specify a sane limit for the converted name.
     */
    nam_len = loc2utf( config->codepage,
                       ce->ce_mode & S_UTF16BE ? T_UTF_16BE : T_UTF_8,
                       ce->ce_name, strlen(ce->ce_name) + 1,
                       te->te_name, -1U );
    config->private += sizeof (struct tree_entry) + nam_len;
    return 0;
}

int
write_tree (struct config *config, struct dindex *dindex, uint8_t *sha1)
{
    char pp_sha1 [41];
    uint8_t buf [20];

    if (!sha1)
        sha1 = buf;
    
    if (write_tree_inner (config, &dindex->header->hdr_root, sha1) < 0)
        return -1;

    info ("Tree %s written.", format_sha1_str (pp_sha1, sha1));
    return 0;
}

/******************************************************************************/

#include <time.h>

static int
daylight_saving (time_t *tv)
{
    struct tm gmt;
    struct tm loc;

    gmtime_r (tv, &gmt), localtime_r (tv, &loc);
    return 100*loc.tm_hour + loc.tm_min - 100*gmt.tm_hour - gmt.tm_min;
}

int
commit_tree (struct config *config, uint8_t *sha1, const uint8_t **parents)
{
    int i, offs;
    time_t now;
    unsigned long tick;
    size_t size, limit;
    char pp_sha1 [41];

    errno = EINVAL;
    if (!sha1)
        return -1;

    /*
     *  Get time informations.
     */
    tick = (unsigned long) time (&now);
    offs = daylight_saving (&now);

    /*
     *  This loop may be executed twice if the initial size (limit)
     *  of the data buffer is not large enough to hold the complete
     *  commit object.  Of course, this is highly unlikely!
     */
    for (limit = 4096; 1; limit = size + 1)
    {
        char commit [limit];

        /*
         *  1st line, referenced tree.
         */
        size = snprintf (commit, limit, "%s %s\n", T_TREE,
                         format_sha1_str (pp_sha1, sha1));

        /*
         *  2nd to (N-1)th line, parents.
         */
        if (parents)
            for (i=0; parents[i]; i++)
                size += snprintf (commit+size, limit > size ? limit-size : 0,
                                  "%s %s\n", T_PARENT,
                                  format_sha1_str (pp_sha1, parents[i]));

        /*
         *  Nth line, machine this was generated on and when.
         */
        size += snprintf (commit+size, limit > size ? limit-size : 0,
                          "%s %s %lu %+05d\n", T_ORIGIN,
                          config->nodename, tick, offs);

        /*
         *  If output buffer was too small, repeat with correct size.
         */
        if (size >= limit)
            continue;

        /*
         *  Finally write the commit file.
         */
        if (write_sha1_file (config->dbase, T_COMMIT, NULL, commit, size, sha1) < 0)
            return -1;
        info ("Commit %s written.", format_sha1_str (pp_sha1, sha1));
        return 0;
    }
}
