/******************************************************************************\
 *                        ____            _ _ 
 *                       |  _ \ ___ _ __ | (_) ___ __ _
 *                       | |_) / _ \ '_ \| | |/ __/ _` |
 *                       |  _ <  __/ |_) | | | (_| (_| |
 *                       |_| \_\___| .__/|_|_|\___\__,_|
 *                                 |_|
 ******************************************************************************
 *                                 
 *  Copyright (c) 2005-2006, Herbert Nachtnebel
 *
 *    based on code from the GIT project at your favorite kernel mirror [1]
 *
 *  Copyright (c) 2005-2006, Linus Torvalds
 *                           Junio C Hamano
 *                           et al.
 *
 *  [1] http://www.kernel.org/pub/software/scm
 *  
 ******************************************************************************
 *                      
 *  Title:    Replica - File Replication and Backup Utility
 *
 *  Version:  0.1
 *
 *  Date:     01-Nov-2005
 *
 *  Author:   Herbert Nachtnebel (Berri)
 *  
 ******************************************************************************
 *  
 *  Change log:
 *  
 *     Date     Rev  Ver                    Description
 *  ----------- --- ----- ---------------------------------------------------
 *  01-Nov-2005 Ber  0.1  Initial implementation
 *  
 ******************************************************************************
 *  
 *  Description:
 *
 *    The Toolset
 *    -----------
 *    
\******************************************************************************/
#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <openssl/sha.h>
#include <zlib.h>

#include "compat.h"
#include "cache.h"
#include "index.h"
#include "utils.h"
#include "token.h"
#include "update.h"


inline static uint32_t
convert_mode (mode_t st_mode)
{
    return S_POSIX_FIELDS & st_mode;
}

inline static uint64_t
convert_time (time_t st_time)
{
    /*
     *  TODO!  Convert to nanoseconds since epoch (64-bit is large enough).
     */
    return st_time;
}

int
update_metadata (struct cache_entry *ce, struct stat *st)
{
    uint64_t ctime = convert_time (st->st_ctime);
    uint64_t mtime = convert_time (st->st_mtime);
    uint32_t dev   = st->st_dev;
    uint32_t ino   = st->st_ino;
    uint32_t mode  = convert_mode (st->st_mode);
    uint32_t uid   = st->st_uid;
    uint32_t gid   = st->st_gid;
    uint32_t rdev  = st->st_rdev;
    uint64_t size  = st->st_size;

    int rv = 0;

    if (S_ISREG(st->st_mode) || S_ISLNK(st->st_mode))
    {

        if (ce->ce_ctime != ctime ||
            ce->ce_mode  != mode  ||
            ce->ce_uid   != uid   ||
            ce->ce_gid   != gid)
        {
            ce->ce_ctime = ctime;
            ce->ce_mode  = mode;
            ce->ce_uid   = uid;
            ce->ce_gid   = gid;
            rv = METADATA_CHANGED;
        }

        if (ce->ce_mtime != mtime ||
            ce->ce_dev   != dev   ||
            ce->ce_ino   != ino   ||
            ce->ce_size  != size)
        {
            ce->ce_mtime = mtime;
            ce->ce_dev   = dev;
            ce->ce_ino   = ino;
            ce->ce_size  = size;
            rv = FILEDATA_CHANGED;
        }
    }
    else
    {
        if (ce->ce_ctime != ctime ||
            ce->ce_mtime != mtime ||
            ce->ce_dev   != dev   ||
            ce->ce_ino   != ino   ||
            ce->ce_mode  != mode  ||
            ce->ce_uid   != uid   ||
            ce->ce_gid   != gid   ||
            ce->ce_rdev  != rdev)
        {
            ce->ce_ctime = ctime;
            ce->ce_mtime = mtime;
            ce->ce_dev   = dev;
            ce->ce_ino   = ino;
            ce->ce_mode  = mode;
            ce->ce_uid   = uid;
            ce->ce_gid   = gid;
            ce->ce_rdev  = rdev;
            memset (ce->ce_sha1, 0, sizeof ce->ce_sha1);
            rv = METADATA_CHANGED;
        }
    }
    return rv;
}

/******************************************************************************/

int
write_sha1_file (const char *dbase, token type, token subtype,
                 const void *buffer, size_t size, uint8_t *sha1)
{
    int rv;
    ssize_t count = subtype ? strlen (subtype) : 0;
    struct fcache *cf;

    size_t ibuf_len = strlen (type) + count + 20;
    unsigned char ibuf [ibuf_len];

    size_t obuf_len = (ibuf_len + 4096) - (ibuf_len + 4096) % 4096;
    unsigned char obuf [obuf_len];

    SHA_CTX  hash = { 0 };
    z_stream strm = { 0 };

    cf = create_temporary_cache_file (dbase);
    if (!cf)
        goto creation_error;

    /*
     *  Initialize compression.
     */
    SHA1_Init (&hash);
    deflateInit (&strm, Z_BEST_COMPRESSION);
    strm.next_out = obuf, strm.avail_out = obuf_len;

    /*
     *  Start compressing with the blob header.
     */
    if (subtype)
        count = sprintf (ibuf, "%s (%s) %zu", type, subtype, size) + 1;
    else
        count = sprintf (ibuf, "%s %zu", type, size) + 1;

    /*
     *  Push input data through compressor.
     */
    SHA1_Update (&hash, ibuf, count);
    strm.next_in = ibuf, strm.avail_in = count;

    do
    {
        rv = deflate (&strm, size ? Z_NO_FLUSH : Z_FINISH);
        if (!strm.avail_out || !size)
        {
            char  *pos = obuf;
            size_t put = strm.next_out - obuf;

            while (put)
            {
                count = write (cf->fd, obuf, put);
                if (count == -1)
                {
                    if (errno != EINTR)
                        goto hopelessly_broken;
                }
                else
                {
                    pos += count;
                    put -= count;
                }
            }
            strm.next_out = obuf, strm.avail_out = obuf_len;
        }
        if (!strm.avail_in && size)
        {
            SHA1_Update (&hash, buffer, size);
            strm.next_in = buffer, strm.avail_in = size;
            size = 0;
        }
    } while (rv == Z_OK);

    /*
     *  Finally move the newly created blob into its place.
     */
    deflateEnd (&strm);
    SHA1_Final (sha1, &hash);
    if (commit_temporary_cache_file (cf, sha1) < 0)
        goto give_up;    
    return 0;

  hopelessly_broken:
    deflateEnd (&strm);
    
  give_up:
    rv = errno;
    drop_cache_file (cf);
    errno = rv;
    
  creation_error:
    memset (sha1, 0, 20);
    return -1;
}

static int
recover_from_errors (int fd, void *buf, size_t len)
{
    /*
     *  TODO!  Binary search to fill as much intact data as possible
     *         into the block.
     */
    return -1;
}

static int
write_blob (struct cache_entry *ce, const char *file, const char *dbase)
{
    int fd, rv;
    ssize_t count;
    struct fcache *cf;

    SHA_CTX  hash = { 0 };
    z_stream strm = { 0 };

    /*
     *  For this limit I will be cruzified!  Ughh, 1MB on stack.
     *  But this is user space!  Any architecture which can't give us
     *  a single megabyte on the stack isn't worth programming for.
     */
    size_t ibuf_len = ce->ce_size > 1024*1024 ? 1024*1024 : 32768;
    unsigned char ibuf [ibuf_len];

    /*
     *  In future, when pushing over LAN multiple of MTU?
     */
    size_t obuf_len = 32768;
    unsigned char obuf [obuf_len];

    fd = open (file, O_RDONLY | O_LARGEFILE | O_NOATIME);
    if (fd < 0)
        goto open_error;

    cf = create_temporary_cache_file (dbase);
    if (!cf)
        goto creation_error;

    /*
     *  Initialize compression.
     */
    SHA1_Init (&hash);
    deflateInit (&strm, Z_BEST_COMPRESSION);
    strm.next_out = obuf, strm.avail_out = obuf_len;

    /*
     *  Start compressing with the blob header.
     *
     *  Note!  The size is only a hint!  If the file has grown in the
     *         time since the stat system call, then more data will be
     *         compressed into the blob. Tough...
     */
    count = sprintf (ibuf, "%s %u", T_BLOB, (unsigned int)ce->ce_size) + 1;

    /*
     *  Push input data through compressor.
     */
    do {
        SHA1_Update (&hash, ibuf, count);
        strm.next_in = ibuf, strm.avail_in = count;

        while ((rv = deflate (&strm, Z_NO_FLUSH)) == Z_OK)
        {
            if (strm.avail_out == 0)
            {
                size_t put = strm.next_out - obuf;
                char *pos = obuf;

                while (put)
                {
                    count = write (cf->fd, obuf, put);
                    if (count == -1)
                    {
                        if (errno != EINTR)
                            goto hopelessly_broken;
                    }
                    else
                    {
                        pos += count;
                        put -= count;
                    }
                }
                strm.next_out = obuf, strm.avail_out = obuf_len;
            }
            if (strm.avail_in == 0)
                break;
        }

        do {
            count = read (fd, ibuf, ibuf_len);
            if (count == -1)
            {
                if (errno == EIO)
                {
                    /*
                     *  Since this application is also used for backup
                     *  means, we try to read as much correct data as
                     *  possible.  Nevertheless we remember that there
                     *  were read errors and this file is broken.
                     */
                    ce->ce_mode |= S_BROKEN_DATA;
                    count = recover_from_errors (fd, ibuf, ibuf_len);
                    if (count == -1)
                        goto hopelessly_broken;
                }
                else if (errno != EINTR)
                    goto hopelessly_broken;
            }            
        } while (count == -1);
    } while (count > 0);

    /*
     *  Finish the compression.
     */
    while (rv == Z_OK)
    {
        size_t put;
        char *pos;
        
        rv = deflate (&strm, Z_FINISH);
        put = strm.next_out - obuf;
        pos = obuf;

        while (put)
        {
            count = write (cf->fd, obuf, put);
            if (count == -1)
            {
                if (errno != EINTR)
                    goto hopelessly_broken;
            }
            else
            {
                pos += count;
                put -= count;
            }
        }
        strm.next_out = obuf, strm.avail_out = obuf_len;
    }
    deflateEnd (&strm);

    /*
     *  Finally move the newly created blob into its place.
     */
    SHA1_Final (ce->ce_sha1, &hash);
    if (commit_temporary_cache_file (cf, ce->ce_sha1) < 0)
        goto give_up;
    
    close (fd);
    return 0;

  hopelessly_broken:
    deflateEnd (&strm);
    
  give_up:
    rv = errno;
    drop_cache_file (cf);
    errno = rv;
    
  creation_error:
    memset (ce->ce_sha1, 0, sizeof (ce->ce_sha1));
    ce->ce_mode |= S_BROKEN_DATA;
    rv = errno;
    close (fd);
    errno = rv;
    return -1;
    
  open_error:
    ce->ce_mode |= (errno == ENOENT) ? S_MISSING_FILE : S_BROKEN_DATA;
    return -1;
}

static int
write_link (struct cache_entry *ce, const char *file, const char *dbase)
{
    int rv;
    ssize_t count;
    struct fcache *cf;

    size_t ibuf_len = ce->ce_size + 4096;
    unsigned char ibuf [ibuf_len];

    size_t obuf_len = (ibuf_len + 4096) - (ibuf_len + 4096) % 4096;
    unsigned char obuf [obuf_len];

    SHA_CTX  hash = { 0 };
    z_stream strm = { 0 };

    cf = create_temporary_cache_file (dbase);
    if (!cf)
        goto creation_error;

    /*
     *  Initialize compression.
     */
    SHA1_Init (&hash);
    deflateInit (&strm, Z_BEST_COMPRESSION);
    strm.next_out = obuf, strm.avail_out = obuf_len;

    /*
     *  Start compressing with the blob header.
     */
    count = sprintf (ibuf, "%s %u", T_BLOB, (unsigned int)ce->ce_size) + 1;
    if ((rv = readlink (file, ibuf + count, ibuf_len - count)) < 0)
        goto hopelessly_broken;
    count += rv;

    /*
     *  Push input data through compressor.
     */
    SHA1_Update (&hash, ibuf, count);
    strm.next_in = ibuf, strm.avail_in = count;
    do {
        size_t put;
        char *pos;
        
        strm.next_out = obuf, strm.avail_out = obuf_len;
        rv = deflate (&strm, Z_FINISH);
        put = strm.next_out - obuf;
        pos = obuf;

        while (put)
        {
            count = write (cf->fd, obuf, put);
            if (count == -1)
            {
                if (errno != EINTR)
                    goto hopelessly_broken;
            }
            else
            {
                pos += count;
                put -= count;
            }
        }
    } while (rv == Z_OK);
    deflateEnd (&strm);

    /*
     *  Finally move the newly created blob into its place.
     */
    SHA1_Final (ce->ce_sha1, &hash);
    if (commit_temporary_cache_file (cf, ce->ce_sha1) < 0)
        goto give_up;
    return 0;

  hopelessly_broken:
    deflateEnd (&strm);
    
  give_up:
    rv = errno;
    drop_cache_file (cf);
    errno = rv;
    
  creation_error:
    memset (ce->ce_sha1, 0, sizeof (ce->ce_sha1));
    ce->ce_mode |= S_BROKEN_DATA;
    return -1;
}

/******************************************************************************/

int
update_filedata (struct cache_entry *ce, const char *file, const char *dbase)
{
    int rv;

    if (S_ISREG(ce->ce_mode))
    {
        rv = write_blob (ce, file, dbase);
    }
    else if (S_ISLNK(ce->ce_mode))
    {
        rv = write_link (ce, file, dbase);
    }
    else
    {
        errno = EINVAL;
        rv = -1;
    }
    return rv;
}

/******************************************************************************/

/*
 *  Search for our data header, interpret the information found there and
 *  strip it from the stream.
 */
static int
parse_header (z_stream *strm, token *type, token *subtype, size_t *size)
{
    /*
     *  The data header can have the following formats:
     *
     *    %s (%s) %u\0
     *    %s %u\0
     *
     *  with the fields "type", "subtype" and "data length".
     */
    int n, s, t;
    char *hdr = strm->next_out - strm->total_out;

    /*
     *  Search for the type field.
     */
    for (n = s = t = 0; n < strm->total_out; t++, s++, n++)
        if (hdr[n] <= ' ')
            break;

    /*
     *  Check the subtype field.
     */
    if (hdr[n+1] == '(')
        for (s++, n++; n < strm->total_out; s++)
            if (hdr[s] < ' ' || hdr[n++] == ')')
                break;

    /*
     *  Length field.
     */
    if (hdr[n] == ' ')
        for (n++; n < strm->total_out; n++)
            if (hdr[n] < '0' || hdr[n] > '9')
                break;

    /*
     *  Ok, check and interpret the header.
     */
    if (hdr[n] == 0 && hdr[t] == ' ')
    {
        hdr[t] = 0;
        *type = unique (hdr);
        *size = atol (hdr+t+1);
        if (hdr[t+1] == '(' && hdr[s] == ')')
        {
            hdr[s] = 0;
            *subtype = unique (hdr+t+2);
            *size = atol (hdr+s+2);
        }

        /*
         *  Strip the header from the stream.
         */
        strm->avail_out += n+1;
        strm->total_out -= n+1;
        strm->next_out  -= n+1;
        memmove (hdr, hdr+n+1, strm->total_out);
        return 0;
    }

    /*
     *  If we are shure that the header is broken, return an error.
     */
    return hdr[n] < ' ' ? -1 : 0;
}

/*
 *  Read-in and interpret a cache file from our database.
 *
 *  Note!  If the decompressed data stream reaches the given limit, then
 *         the decompression is aborted.  If the file header could be found
 *         in the data stream up to this limit, then all header informations
 *         are nevertheless put into the return arguments.
 */
int
read_sha1_file (const char *dbase, const uint8_t *sha1, size_t limit,
                token *type, token *subtype, size_t *size, void **buffer)
{
    int fd, rv = Z_OK;

    SHA_CTX  hash = { 0 };
    z_stream strm = { 0 };

    ssize_t got;
    uint8_t chk_sha1 [20];

    size_t ibuf_len = 32768;
    char ibuf [ibuf_len];

    size_t obuf_len = 0;
    char *obuf = NULL;
    
    fd = open_cache_file_for_reading (dbase, sha1);
    if (fd < 0)
        return -1;

    SHA1_Init (&hash);
    inflateInit (&strm);
    *type = *subtype = NULL, *buffer = NULL, *size = 0;
    if (limit == 0)
        limit = -1UL;

    /*
     *  Decompress the data stream.
     */
    while (1)
    {
        got = read (fd, ibuf, ibuf_len);
        if (got == -1)
        {
            if (errno == EINTR)
                continue;
            goto data_error;
        }

        errno = EINVAL;
        if (got == 0)
        {
            if (strm.avail_in == 0)
                break;
            goto data_error;
        }
        if (rv == Z_STREAM_END)
            goto data_error;
        
        strm.avail_in = got;
        strm.next_in = ibuf;
        while (strm.avail_in != 0 && rv != Z_STREAM_END)
        {                
            /*
             *  Grow the output buffer.
             */
            if (strm.avail_out == 0)
            {
                size_t upper_bound = *size < limit ? *size : limit;

                obuf_len = obuf_len < upper_bound ? upper_bound :
                           obuf_len + 4096 - obuf_len % 4096;
                obuf = realloc (obuf, obuf_len);
                errno = ENOMEM;
                if (!obuf)
                    goto data_error;
                strm.avail_out = obuf_len - strm.total_out;
                strm.next_out = obuf + strm.total_out;
            }

            /*
             *  Pump input data through inflate.
             */
            got = strm.avail_out;
            rv = inflate (&strm, Z_NO_FLUSH);
            errno = EIO;
            if (rv != Z_OK && rv != Z_STREAM_END)
                goto data_error;
            got -= strm.avail_out;
            SHA1_Update (&hash, strm.next_out - got, got);

            /*
             *  If we haven't seen the header, try to find it.
             */
            errno = EINVAL;
            if (*type == NULL && strm.total_out != 0)
                if (parse_header (&strm, type, subtype, size) < 0)
                    goto data_error;

            if (strm.total_out >= limit)
                goto enforce_limit;
        }
    }

    /*
     *  After checking the SHA1 of the decompressed stream, return it.
     */
    *size = strm.total_out;
    SHA1_Final (chk_sha1, &hash);
    errno = EINVAL;
    if (memcmp (chk_sha1, sha1, sizeof chk_sha1) != 0)
        goto data_error;

  enforce_limit:
    *buffer = obuf;
    inflateEnd (&strm);
    return close (fd);

  data_error:
    rv = errno;
    inflateEnd (&strm);
    free (obuf);
    close (fd);
    errno = rv;
    return -1;
}

/******************************************************************************/

void
dump_type_of (const char *dbase, const char *argv)
{
    if (dbase && argv)
    {
        uint8_t sha1 [20];

        if (string_to_sha1 (sha1, argv))
        {
            token type, subtype;
            size_t size;
            void *buffer;

            if (read_sha1_file (dbase, sha1, 8192,
                                &type, &subtype, &size, &buffer) < 0)
                die ("Error reading sha1 file: %s", strerror(errno));

            if (subtype)
                printf ("type=%s subtype=%s size=%zu\n", type, subtype, size);
            else
                printf ("type=%s size=%zu\n", type, size);
            free (buffer);
        }
        else
            die ("Invalid sha1: %s", argv);
    }
}

void
dump_sha1_cat (const char *dbase, const char *argv)
{
    if (dbase && argv)
    {
        uint8_t sha1 [20];

        if (string_to_sha1 (sha1, argv))
        {
            token type, subtype;
            size_t size;
            void *buffer;

            if (read_sha1_file (dbase, sha1, 0,
                                &type, &subtype, &size, &buffer) < 0)
                die ("Error reading sha1 file: %s", strerror(errno));

            fwrite (buffer, size, 1, stdout);
            fflush (stdout);
            free (buffer);
        }
        else
            die ("Invalid sha1: %s", argv);
    }
}

