/******************************************************************************\
 *                        ____            _ _ 
 *                       |  _ \ ___ _ __ | (_) ___ __ _
 *                       | |_) / _ \ '_ \| | |/ __/ _` |
 *                       |  _ <  __/ |_) | | | (_| (_| |
 *                       |_| \_\___| .__/|_|_|\___\__,_|
 *                                 |_|
 ******************************************************************************
 *                                 
 *  Copyright (c) 2005, Linus Torvalds
 *                      Herbert Nachtnebel
 *                      
 ******************************************************************************
 *                      
 *  Title:    Replica - File Replication and Backup Utility
 *
 *  Version:  0.1
 *
 *  Date:     01-Nov-2005
 *
 *  Author:   Herbert Nachtnebel (Berri)
 *  
 ******************************************************************************
 *  
 *  Change log:
 *  
 *     Date     Rev  Ver                    Description
 *  ----------- --- ----- ---------------------------------------------------
 *  01-Nov-2005 Ber  0.1  Initial implementation
 *  
 ******************************************************************************
 *  
 *  Description:
 *
 *    The Toolchest
 *    -------------
 *    
\******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <memory.h>
#include <errno.h>

#include "compat.h"
#include "utils.h"
#include "token.h"

void die (const char *msg, ...)
{
    va_list args;

    va_start (args, msg);
    vfprintf (stderr, msg, args);
    fputc ('\n', stderr);
    va_end (args);
    exit (1);
}

void info (const char *msg, ...)
{
    va_list args;

    va_start (args, msg);
    vfprintf (stderr, msg, args);
    fputc ('\n', stderr);
    va_end (args);
}

/******************************************************************************/

static const char hex[] = "0123456789abcdef";

char*
fill_sha1_path (char *path, const uint8_t *sha1)
{
    int i;
    char *pos = path;

    for (i = 0; i < 20; i++)
    {
        unsigned int val = sha1[i];

        while (*pos != 'X') pos++;
        *pos++ = hex[val >> 4];
        while (*pos != 'X') pos++;
        *pos++ = hex[val & 0xf];
    }
    return path;
}

char*
format_sha1_str (char *str, const uint8_t *sha1)
{
    int i;
    char *pos = str;

    for (i = 0; i < 20; i++)
    {
        unsigned int val = sha1[i];

        *pos++ = hex[val >> 4];
        *pos++ = hex[val & 0xf];
    }
    *pos++ = 0;
    return str;
}

static unsigned int
conv_digit (unsigned int ch)
{
    ch -= '0';
    if (ch < 10)
        return ch;
    ch += '0' - 'A';
    if (ch < 26)
        return ch + 10;
    ch += 'A' - 'a';
    if (ch < 26)
        return ch + 10;
    return -1U;
}

uint8_t *
string_to_sha1 (uint8_t *sha1, const char *str)
{
    int i;
    unsigned char value = 0;

    /*
     *  We expect 40 hex characters...
     */
    for (i = 0; i < 40; i++)
    {
        unsigned int digit = conv_digit (str[i]);

        if (digit < 16)
            value = digit + (value << 4);
        else
            goto error;
        if (i & 1)
            sha1[i>>1] = value;
    }

    /*
     *  and then the string end.
     */
    if (str[i] != 0)
    {
      error:
        memset (sha1, 0, 20);
        sha1 = NULL;
    }
    return sha1;
}

/******************************************************************************/

/*
 *  Note!  All four UTF conversion routines currently untested!
 */

int
utf8_encode (char *dst, int ucs4)
{
    if (ucs4 >= 0)
    {
        if (ucs4 < 0x00000080)
        {
            dst[0] = 0x80;
            return 1;
        }
        if (ucs4 < 0x00000800)
        {
            dst[1] = 0x80 + (ucs4 & 0x3F);  ucs4 >>= 6;
            dst[0] = 0xC0 + (ucs4 & 0x1F);
            return 2;
        }
        if (ucs4 < 0x00010000)
        {
            dst[2] = 0x80 + (ucs4 & 0x3F);  ucs4 >>= 6;
            dst[1] = 0x80 + (ucs4 & 0x3F);  ucs4 >>= 6;
            dst[0] = 0xE0 + (ucs4 & 0x0F);
            return 3;
        }
        if (ucs4 < 0x00200000)
        {
            dst[3] = 0x80 + (ucs4 & 0x3F);  ucs4 >>= 6;
            dst[2] = 0x80 + (ucs4 & 0x3F);  ucs4 >>= 6;
            dst[1] = 0x80 + (ucs4 & 0x3F);  ucs4 >>= 6;
            dst[0] = 0xF0 + (ucs4 & 0x07);
            return 4;
        }
        /*
         *  Forget about higher order encodings since
         *  UTF-16 is also not able to code them.
         */
    }
    return 0;
}

int
utf8_decode (char **src)
{
    unsigned char val;
    int rv, state;

    rv = state = 0;
    while (1) switch (((val = **src) & 0xF8) >> 3)
    {
        case 0 ... 15:
            if (state)
                break;
            (*src)++;
            return val;
                
        case 16 ... 23:
            (*src)++;
            if (state)
            {
                rv = (rv << 6) + val - 0x80;
                if (--state)
                    continue;
                return rv;
            }
            rv = val;
            break;

        case 24 ... 27:
            if (state)
                break;
            (*src)++;
            rv = val - 0xC0;
            state = 1;
            continue;

        case 28 ... 29:
            if (state)
                break;
            (*src)++;
            rv = val - 0xE0;
            state = 2;
            continue;

        case 30:
            if (state)
                break;
            (*src)++;
            rv = val - 0xF0;
            state = 3;
            continue;

        case 31:
            if (state)
                break;
            (*src)++;
            rv = val;
            break;                
    }
    return rv | 0x80000000;
}

int
utf16_encode (short *dst, int ucs4)
{
    if (ucs4 >= 0)
    {
        if (ucs4 < 0x00010000)
        {
            dst[0] = ucs4;
            return 1;
        }
        if (ucs4 < 0x00110000)
        {
            dst[1] = 0xDC00 + (ucs4 & 0x3FF); ucs4 >>= 10;
            dst[0] = 0xD800 + (ucs4 & 0x3FF);
            return 2;
        }
    }
    return 0;
}

int
utf16_decode (short **src)
{
    unsigned short val;
    int rv;

    rv = val = *(*src)++;
    if (val >= 0xD800 && val <= 0xDFFF)
    {
        rv = val - 0xD800;
        if (val >= 0xDC00)
            return rv | 0x80000000;
        
        val = *(*src)++;
        if (val >= 0xD800 && val <= 0xDFFF)
        {
            rv = (rv << 10) + (val - 0xDC00);
            if (val < 0xDC00)
                return rv | 0x80000000;
        }
    }
    return rv;
}

/******************************************************************************/

#include <iconv.h>

size_t
remove_escape_sequencies (char *utf16, size_t len)
{
    /*
     *  TODO:  Undo U+xxxx escape sequencies.
     */
    return len;
}

size_t
utf16be_to_utf8 (const char *utf16, size_t utf16_len,
                 char *utf8, size_t utf8_len)
{
    static iconv_t conv = (iconv_t) 0;
    size_t buf_len = utf16_len;

    if (conv == (iconv_t) 0)
        conv = iconv_open (T_UTF_8, T_UTF_16BE);
    if (conv == (iconv_t) -1)
        die ("Can't initialize conversion from %s to %s", T_UTF_16BE, T_UTF_8);

    for (buf_len = buf_len + 256 - buf_len % 256; 1; buf_len += buf_len)
    {
        char *src, *dst;
        size_t src_len, dst_len;

        char tmp_buf [buf_len];

        src = utf16,   src_len = utf16_len;
        dst = tmp_buf, dst_len = buf_len;

        if (-1 == iconv (conv, &src, &src_len, &dst, &dst_len))
        {
            if (errno == E2BIG)
            {
                /*
                 *  Reset conversion and start over with a larger buffer.
                 */
                iconv (conv, NULL, 0, NULL, 0);
                continue;
            }
            /*
             *  TODO!  Give hints about the filename.
             */
            die ("You have an unicode sequence error in a filename");
        }

        dst_len = buf_len - dst_len;
        if (utf8_len > 0)
            memcpy (utf8, tmp_buf, utf8_len < dst_len ? utf8_len : dst_len);
        return dst_len;
    }
}

size_t
loc2utf (const char *codepage, const void *utf_variant,
         const char *loc, size_t loc_len, char *utf, size_t utf_len)
{
    /*
     *  Note!  As long as the codepage parameter isn't replaced with an opaque
     *         type describing the conversion, just create a conversion handle
     *         every time.
     */
    size_t buf_len = 2 * loc_len;
    iconv_t conv = iconv_open (T_UTF_16BE, codepage);

    if (conv == (iconv_t) -1)
        die ("Can't initialize conversion from %s to %s", codepage, T_UTF_16BE);

    for (buf_len = buf_len + 256 - buf_len % 256; 1; buf_len += buf_len)
    {
        char *src, *dst;
        size_t src_len, dst_len;

        char tmp_buf [buf_len];

        src = loc,     src_len = loc_len;
        dst = tmp_buf, dst_len = buf_len;

        if (-1 == iconv (conv, &src, &src_len, &dst, &dst_len))
        {
            if (errno == E2BIG)
            {
                /*
                 *  Reset conversion and start over with a larger buffer.
                 */
                iconv (conv, NULL, 0, NULL, 0);
                continue;
            }

            /*
             *  Ughh, now we have a problem.  There apears to be a filename
             *  which can't be translated into Unicode?!  That's weird!
             *  I don't know how to handle that well, so let's inform
             *  the user and just die.
             */
            die ("You have an invalid filename in your system %s", loc);
        }
        iconv_close (conv);

        /*
         *  If a filename originally was created on another system with
         *  another coding, then it might be possible that non-mappable
         *  characters were translated into U+xxxx escape sequencies.
         *  Since we are going back now and re-transcribe the name to
         *  Unicode, undo all such transliterations.
         */
        buf_len = remove_escape_sequencies (tmp_buf, buf_len - dst_len);

        /*
         *  Now transcode into UTF-8.  It directly write's to the output
         *  buffer under the assumption, that most of the time it is 
         *  more efficient to use this encoding.  This assumption will
         *  break for east asian countries, but we are going to cure it
         *  below.
         */
        if (utf_variant != T_UTF_16BE)
            dst_len = utf16be_to_utf8 (tmp_buf, buf_len, utf, utf_len);
        if (utf_variant == T_UTF_8)
            return dst_len;

        /*
         *  If we have a destination buffer, then look whether the original
         *  UTF-16 resulted in a more efficient coding and use it in case
         *  it does.
         */
        if (utf_len > 0 && (utf_variant == T_UTF_16BE || dst_len > buf_len))
            memcpy (utf, tmp_buf, utf_len < buf_len ? utf_len : buf_len);
        if (utf_variant == T_UTF_16BE)
            return buf_len;

        /*
         *  When estimating the needed buffer length, tell the caller
         *  whether the UTF-8 encoding or the UTF-16 is more efficient
         *  in terms of memory usage.
         */        
        if (dst_len > buf_len)
        {
            *(token*)utf_variant = T_UTF_16BE;
            return buf_len;
        }

        *(token*)utf_variant = T_UTF_8;
        return dst_len;
    }
}

size_t
utf2loc (const char *codepage,
         const char *utf, size_t utf_len, char *loc, size_t loc_len)
{
    return 0;
}
