/******************************************************************************\
 *                        ____            _ _ 
 *                       |  _ \ ___ _ __ | (_) ___ __ _
 *                       | |_) / _ \ '_ \| | |/ __/ _` |
 *                       |  _ <  __/ |_) | | | (_| (_| |
 *                       |_| \_\___| .__/|_|_|\___\__,_|
 *                                 |_|
 ******************************************************************************
 *                                 
 *  Copyright (c) 2005-2006, Herbert Nachtnebel
 *                      
 ******************************************************************************
 *                      
 *  Title:    Replica - File Replication and Backup Utility
 *
 *  Version:  0.1
 *
 *  Date:     10-Mar-2006
 *
 *  Author:   Herbert Nachtnebel (Berri)
 *  
 ******************************************************************************
 *  
 *  Change log:
 *  
 *     Date     Rev  Ver                    Description
 *  ----------- --- ----- ---------------------------------------------------
 *  10-Mar-2006 Ber  0.1  Initial implementation
 *  
 ******************************************************************************
 *  
 *  Description:
 *
 *    Tokens are strings stored uniquely in a binary sorted array.
 *
 *    Note!  Three reasons to use them:
 *
 *           1) They make the code more readable since many string compares
 *              can be replaced with pointer compares;  the pointer is
 *              equivalent to the actual string contents.
 *
 *           2) They avoid strdup's and, successively, memory management when
 *              returning strings from functions.
 *
 *           3) When looking up new tokens is seldom and comparision is done
 *              often, the application will run faster since strcmp's can be
 *              replaced with more efficient pointer compares.  That's a huge
 *              advantage for interpreted languages like a lisp interpreter
 *              or even for a simple parser.
 *    
\******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "token.h"

/******************************************************************************/

/*
 *  String constants used in this application.
 *
 *  Note!  Keeping them sorted alphabetically is advisable, since it minimizes
 *         the application startup time - building the uniquiness database is
 *         then O(N).  But it is _not_ required for correctness.
 */
TOKEN T_BLOB     = "blob";
TOKEN T_COMMIT   = "commit";
TOKEN T_ORIGIN   = "origin";
TOKEN T_PARENT   = "parent";
TOKEN T_POSIX    = "posix";
TOKEN T_TREE     = "tree";
TOKEN T_UTF_16BE = "UTF-16BE";
TOKEN T_UTF_8    = "UTF-8";

/******************************************************************************/

/*
 *  unique
 *  ------
 *  
 *    Create a unique token (string object).
 *
 *    Tokens are stored only once, to assure this, a sorted list of tokens
 *    is kept.
 *
 *    Note!  Super simple implementation of a binary sorted array.  Insertion
 *           has bad O(N) worst case runtime and building the complete database
 *           thus O(N), but it is seldom run and barely runtime critical.
 *           Hence, not worth for optimization.  On the other hand, the already
 *           used avl tree implementation would be an easy exploitable
 *           replacement.  Maybe...
 */
token
unique (const char *s)
{
    extern const char __bss_start;
    
    static token *xref = 0;
    static size_t xtop = 0;

    int c = 0;
    unsigned int n = 0, l = 0, u = xtop;

    /* binary search for an existing token */
    while (l != u)
    {
        n = (l+u)/2;
        c = strcmp (xref[n], s);
        if (c < 0) u = n; else if (c > 0) l = n+1; else
        {
            return xref[n];
        }
    }

    /* create a new token at the position given by l (or u) */
    if (xtop % 32 == 0)
        xref = realloc (xref, (xtop + 32) * sizeof *xref);

    for (n = xtop++; n > l; n--)
        xref[n] = xref[n-1];

    /*
     * Note!  Attention, linker trick!  Don't duplicate string constants.
     *        Works, since the linker places all constant data before the
     *        __bss_start symbol (since they are initialized).
     */
    xref[l] = s < &__bss_start ? s : strdup(s);
    return xref[l];
}

/*
 *  init_tokens
 *  -----------
 *
 *    Initialize all tokens found in this application and declared as such.
 *
 *    Note!  Linker trick II.  All tokens have to be put into the ".token"
 *           section.  Then, with a linker script, this section is placed
 *           into the read-only data section and is surrounded by the
 *           __beg_tokens and __end_tokens labels.  Therefore, all pointers
 *           found in between can easily be put into the uniqueness database.
 */
void
init_tokens (void)
{
    extern token __beg_tokens;
    extern token __end_tokens;
    token *pt;

    for (pt = &__beg_tokens; pt < &__end_tokens; pt++)
        if (*pt)
            unique(*pt);
}

/*
 *  Execute the init_tokens function before main is run!
 *
 *  Note!  Linker trick III.  The init_tokens function should be run with
 *         all initializers of the application.  This is achieved by putting
 *         a pointer to the function into the ".ctors" section.
 */
INIT_FUNCTION (init_tokens)

